from .are_lines_equal import are_lines_equal
from .main import parse_points_to_contour
from .main import parse_points_to_inner_contour
from .simplify_line import simplify_line
from .compress_polygon_points import compress_polygon_points
from .overlap_object import overlap_object



from ...shape_fitting.fit_rect_ellipse import fit_rect_around_ellipse


def json2bbox(scf_object: dict) -> tuple:
    """
    Convert a standard AxionBio format object (polygon, ellipse, or circle) into bounding box coordinates
    (x_min, y_min, x_max, y_max). All bounding box values are rounded using Python's internal 'round' function
    (https://wiki.c2.com/?BankersRounding).

    Args:
        scf_object (Dict[str, any]): A single annotation in standard AxionBio format. Allowed types are polygon, ellipse, or circle.

    Returns:
        Tuple[int, int, int, int]: Bounding box coordinates in the format (x_min, y_min, x_max, y_max).

    Note:
        - The angle of rotation for an ellipse is taken into account.

    Raises:
        ValueError: If the type of the object is not supported.

    Example::

        object = {
            "type": "polygon",
            "points": [{'x': 0, 'y': 0}, {'x': 20, 'y': 0}, {'x': 20, 'y': 30}, {'x': 0, 'y': 30}]
        }
        bbox = json2bbox(object)
        print(bbox)
        (0, 0, 20, 30)

    """

    if "type" in scf_object:
        if scf_object["type"] == "polygon":
            x_values, y_values = [val["x"] for val in scf_object["points"]], [
                val["y"] for val in scf_object["points"]
            ]
            return (
                int(round(min(x_values))),
                int(round(min(y_values))),
                int(round(max(x_values))),
                int(round(max(y_values))),
            )
        elif scf_object["type"] in ["circle", "ellipse"]:
            x1, y1, x2, y2 = fit_rect_around_ellipse(
                scf_object["center"]["x"],
                scf_object["center"]["y"],
                scf_object["radiusX"],
                scf_object["radiusY"],
                scf_object["angleOfRotation"],
            )
            return x1, y1, x2, y2
        else:
            raise ValueError(f"{scf_object['type']} is not a supported type")

    else:
        raise KeyError(
            f"The scf_object does not have the key 'type'. dict is in the wrong format. {scf_object}"
        )



import numpy as np


def json2contours(json: dict) -> np.ndarray:
    """
    Convert a JSON object in a default AxionBio format to a contour represented as a NumPy array.

    Args:
        json (dict): A JSON object containing 'points' in the form of a list of dictionaries
                     with keys 'x' and 'y', and 'type' set to 'polygon'.

    Returns:
        np.ndarray: A contour represented as a NumPy array.

    Raises:
        ValueError: If the 'type' in the JSON object is not 'polygon' or missing.
    """

    if json.get("type", "").lower() == "polygon":
        result = [[[i["x"], i["y"]]] for i in json["points"]]
    else:
        raise ValueError(
            f"The type {json.get('type', 'NO TYPE GIVEN')} is not supported"
        )
    return np.array(result, dtype=np.int32)



import uuid
import numpy as np
from typing import List, Union


def contours2json(contours: Union[np.ndarray, List[list]]) -> List[dict]:
    """
    Convert a list or array of contours into a list of JSON objects in standard AxionBio format.

    Args:
        contours (numpy.ndarray or list): Contours produced by OpenCV, represented as a NumPy array or a list.

    Returns:
        List[dict]: A list of JSON objects, where each JSON object represents a contour as a polygon.

    """
    result = []
    for contour in contours:
        shape = {"type": "polygon", "points": []}
        for point in contour:
            shape["points"].append(dict(x=int(point[0][0]), y=int(point[0][1])))
        shape["id"] = str(uuid.uuid4())
        shape["parents"] = []
        shape["children"] = []
        result.append(shape)
    return result



from unittest import TestCase

import cv2
import numpy as np

from tomni.annotation_manager import Ellipse, Point, Polygon
from tomni.annotation_manager.main import AnnotationManager


def assertAlmostEqualDictEqual(testcase, expected, actual, percentage_tolerance):
    testcase.assertIsInstance(expected, dict)
    testcase.assertIsInstance(actual, dict)

    testcase.assertSetEqual(set(expected.keys()), set(actual.keys()))

    for key in expected:
        if isinstance(expected[key], dict) and isinstance(actual[key], dict):
            assertAlmostEqualDictEqual(
                testcase, expected[key], actual[key], percentage_tolerance
            )
        elif isinstance(expected[key], (int, float)) and isinstance(
            actual[key], (int, float)
        ):
            if expected[key] == 0:
                testcase.assertAlmostEqual(expected[key], actual[key], places=0)
            else:
                percentage_diff = (
                    abs((expected[key] - actual[key]) / expected[key]) * 100
                )
                testcase.assertLessEqual(
                    percentage_diff,
                    percentage_tolerance,
                    msg=f"Percentage difference for {key}: {percentage_diff:.2f}%",
                )
        else:
            testcase.assertEqual(expected[key], actual[key])


class TestEllipse(TestCase):
    def setUp(self) -> None:
        id_ = "1234-1234-2134-1321"
        children = []
        parents = []
        label = "ellipse_test"

        self.zero_ellipse = Ellipse(
            radius_x=0,
            radius_y=0,
            center=Point(0, 0),
            rotation=0,
            id=id_,
            children=children,
            parents=parents,
            label=label,
        )
        self.circle = Ellipse(
            radius_x=1,
            radius_y=1,
            center=Point(0, 0),
            rotation=0,
            id=id_,
            children=children,
            parents=parents,
            label=label,
        )
        self.oval = Ellipse(
            radius_x=1,
            radius_y=3,
            center=Point(0, 0),
            rotation=0,
            id=id_,
            children=children,
            parents=parents,
            label=label,
        )
        # self.circle_features = Ellipse(radius_x=1, radius_y=1, center=Point(0, 0), rotation=0, id=id_, children=children, parents=parents, label=label, features=["area", "minor_axis", "perimeter"])

    def test_zero_area(self):
        expected = 0.0
        actual = self.zero_ellipse.area

        self.assertEqual(expected, actual)

    def test_circle_area(self):
        expected = np.pi
        actual = self.circle.area

        self.assertAlmostEqual(expected, actual)

    def test_oval_area(self):
        expected = 9.42477796076938
        actual = self.oval.area

        self.assertEqual(expected, actual)

    def test_zero_circularity(self):
        self.assertTrue(np.isnan(self.zero_ellipse.circularity))

    def test_circle_circularity(self):
        expected = 1.0
        actual = self.circle.circularity

        self.assertEqual(expected, actual)

    def test_oval_circularity(self):
        expected = 0.5999999999999999
        actual = self.oval.circularity

        self.assertEqual(expected, actual)

    def test_zero_aspect_ratio(self):
        expected = 1.0
        actual = self.zero_ellipse.aspect_ratio

        self.assertEqual(expected, actual)

    def test_circle_aspect_ratio(self):
        expected = 1.0
        actual = self.circle.aspect_ratio

        self.assertEqual(expected, actual)

    def test_oval_aspect_ratio(self):
        expected = 0.3333333333333333
        actual = self.oval.aspect_ratio

        self.assertEqual(expected, actual)

    def test_zero_perimeter(self):
        expected = 0.0
        actual = self.zero_ellipse.perimeter

        self.assertEqual(expected, actual)

    def test_circle_perimeter(self):
        expected = 6.283185307179586
        actual = self.circle.perimeter

        self.assertEqual(expected, actual)

    def test_oval_perimeter(self):
        expected = 14.049629462081453
        actual = self.oval.perimeter
        self.assertEqual(expected, actual)

    def test_zero_average_diameter(self):
        expected = 0
        actual = self.zero_ellipse.average_diameter
        self.assertEqual(expected, actual)

    def test_circle_average_diameter(self):
        expected = 2
        actual = self.circle.average_diameter
        self.assertEqual(expected, actual)

    def test_oval_average_diameter(self):
        expected = 4
        actual = self.oval.average_diameter
        self.assertEqual(expected, actual)

    def test_zero_minor_axis(self):
        expected = 0
        actual = self.zero_ellipse.minor_axis
        self.assertEqual(expected, actual)

    def test_circle_minor_axis(self):
        expected = 2
        actual = self.circle.minor_axis
        self.assertEqual(expected, actual)

    def test_oval_minor_axis(self):
        expected = 2
        actual = self.oval.minor_axis
        self.assertEqual(expected, actual)

    def test_zero_major_axis(self):
        expected = 0
        actual = self.zero_ellipse.major_axis
        self.assertEqual(expected, actual)

    def test_circle_major_axis(self):
        expected = 2
        actual = self.circle.major_axis
        self.assertEqual(expected, actual)

    def test_oval_major_axis(self):
        expected = 6
        actual = self.oval.major_axis
        self.assertEqual(expected, actual)

    def test_zero_convex_hull_area(self):
        expected = 0.0
        actual = self.zero_ellipse.convex_hull_area
        self.assertEqual(expected, actual)

    def test_circle_convex_hull_area(self):
        expected = np.pi
        actual = self.circle.convex_hull_area
        self.assertEqual(expected, actual)

    def test_oval_convex_hull_area(self):
        expected = 9.42477796076938
        actual = self.oval.convex_hull_area
        self.assertEqual(expected, actual)

    def test_circle_roundness(self):
        expected = 1
        actual = self.circle.roundness
        self.assertEqual(expected, actual)

    def test_oval_roundness(self):
        expected = 1 / 3
        actual = self.oval.roundness
        self.assertEqual(expected, actual)

    def test_zero_ellipse_roundness(self):
        expected = 0
        actual = self.zero_ellipse.roundness
        self.assertEqual(expected, actual)

    def test_zero_ellipse_to_dict(self):
        expected = {
            "area": 0.0,
            "aspectRatio": 1.0,
            "averageDiameter": 0,
            "center": {"x": 0, "y": 0},
            "children": [],
            "circularity": np.nan,
            "convexHullArea": 0.0,
            "id": "1234-1234-2134-1321",
            "label": "ellipse_test",
            "majorAxis": 0,
            "minorAxis": 0,
            "parents": [],
            "perimeter": 0.0,
            "roundness": 0,
            "radiusX": 0.0,
            "radiusY": 0.0,
            "angleOfRotation": 0,
            "type": "ellipse",
            "accuracy": 1,
        }
        actual = self.zero_ellipse.to_dict()

        # Have to use np testing because of verifying nan values.
        np.testing.assert_equal(expected, actual)

    def test_circle_to_dict(self):
        expected = {
            "area": 3.14,
            "aspectRatio": 1.0,
            "averageDiameter": 2,
            "center": {"x": 0, "y": 0},
            "children": [],
            "circularity": 1.0,
            "convexHullArea": 3.14,
            "id": "1234-1234-2134-1321",
            "label": "ellipse_test",
            "majorAxis": 2,
            "minorAxis": 2,
            "parents": [],
            "perimeter": 6.28,
            "radiusX": 1,
            "radiusY": 1,
            "angleOfRotation": 0,
            "type": "ellipse",
            "accuracy": 1,
            "roundness": 1.0,
        }
        actual = self.circle.to_dict()
        self.assertDictEqual(expected, actual)

    def test_oval_to_dict(self):
        expected = {
            "area": 9.42,
            "aspectRatio": 0.33,
            "averageDiameter": 4,
            "center": {"x": 0, "y": 0},
            "children": [],
            "circularity": 0.6,
            "convexHullArea": 9.42,
            "id": "1234-1234-2134-1321",
            "label": "ellipse_test",
            "majorAxis": 6,
            "minorAxis": 2,
            "parents": [],
            "perimeter": 14.05,
            "radiusX": 1,
            "radiusY": 3,
            "angleOfRotation": 0,
            "type": "ellipse",
            "accuracy": 1,
            "roundness": 0.33,
        }
        actual = self.oval.to_dict()

        self.assertDictEqual(expected, actual)

    def test_oval_to_dict_accuracy_05(self):
        expected = {
            "area": 9.42,
            "aspectRatio": 0.33,
            "averageDiameter": 4,
            "center": {"x": 0, "y": 0},
            "children": [],
            "circularity": 0.6,
            "convexHullArea": 9.42,
            "id": "",
            "label": "",
            "majorAxis": 6,
            "minorAxis": 2,
            "parents": [],
            "perimeter": 14.05,
            "radiusX": 1,
            "radiusY": 3,
            "angleOfRotation": 0,
            "type": "ellipse",
            "accuracy": 0.5,
            "roundness": 0.33,
        }
        oval1 = Ellipse(
            radius_x=1,
            radius_y=3,
            center=Point(0, 0),
            rotation=0,
            id="",
            children=[],
            parents=[],
            label="",
            accuracy=0.5,
        )
        actual = oval1.to_dict()

        self.assertDictEqual(expected, actual)

    def test_oval_to_dict_accuracy_0(self):
        expected = {
            "area": 9.42,
            "aspectRatio": 0.33,
            "averageDiameter": 4,
            "center": {"x": 0, "y": 0},
            "children": [],
            "circularity": 0.6,
            "convexHullArea": 9.42,
            "id": "",
            "label": "",
            "majorAxis": 6,
            "minorAxis": 2,
            "parents": [],
            "perimeter": 14.05,
            "radiusX": 1,
            "radiusY": 3,
            "angleOfRotation": 0,
            "type": "ellipse",
            "accuracy": 0,
            "roundness": 0.33,
        }
        oval1 = Ellipse(
            radius_x=1,
            radius_y=3,
            center=Point(0, 0),
            rotation=0,
            id="",
            children=[],
            parents=[],
            label="",
            accuracy=0,
        )
        actual = oval1.to_dict()

        self.assertDictEqual(expected, actual)

    def test_to_dict_with_ellipse_mask(self):
        size = 2072
        quadrant = int(size / 4)
        center = int(size / 2)
        rad = int(size / 3)

        mask_json = AnnotationManager(
            [
                Ellipse(
                    center=Point(center, center),
                    radius_x=rad,
                    rotation=0,
                    id="",
                    label="",
                    children=[],
                    parents=[],
                )
            ]
        ).to_dict()
        ellipse1 = Ellipse(
            radius_x=quadrant,
            center=Point(center, center),
            rotation=0,
            id="",
            label="",
            children=[],
            parents=[],
        )
        ellipse2 = Ellipse(
            radius_x=quadrant - 100,
            center=Point(center + 100, center - 100),
            rotation=0,
            id="",
            label="",
            children=[],
            parents=[],
        )
        ellipse3 = Ellipse(
            radius_x=quadrant - 150,
            center=Point(center - 100, center + 150),
            rotation=0,
            id="",
            label="",
            children=[],
            parents=[],
        )

        ellipses_inside = [ellipse1, ellipse2, ellipse3]
        ellipses_outside = [self.circle, self.oval]

        for ellipse in ellipses_inside:
            self.assertTrue(ellipse.is_in_mask(mask_json=mask_json, min_overlap=0.9))

        for ellipse in ellipses_outside:
            self.assertFalse(ellipse.is_in_mask(mask_json=mask_json, min_overlap=0.9))

    def test_to_dict_with_polygon_mask(self):
        center = int(2072 / 2)
        size = 2072
        quadrant = int(size / 4)
        points = [
            Point(quadrant, quadrant),
            Point(quadrant, quadrant * 3),
            Point(quadrant * 3, quadrant * 3),
            Point(quadrant * 3, quadrant),
            Point(quadrant * 3, quadrant + 1),
        ]
        mask = AnnotationManager(
            [Polygon(points=points, id="", label="", children=[], parents=[])]
        ).to_dict()

        ellipse1 = Ellipse(
            radius_x=quadrant,
            center=Point(center, center),
            rotation=0,
            id="",
            label="",
            children=[],
            parents=[],
        )
        ellipse2 = Ellipse(
            radius_x=quadrant - 100,
            center=Point(center + 100, center - 100),
            rotation=0,
            id="",
            label="",
            children=[],
            parents=[],
        )
        ellipse3 = Ellipse(
            radius_x=quadrant - 150,
            center=Point(center - 100, center + 150),
            rotation=0,
            id="",
            label="",
            children=[],
            parents=[],
        )

        ellipses_inside = [ellipse1, ellipse2, ellipse3]
        ellipses_outside = [self.circle, self.oval]

        for ellipse in ellipses_inside:
            self.assertTrue(ellipse.is_in_mask(mask, 0.9))

        for ellipse in ellipses_outside:
            self.assertFalse(ellipse.is_in_mask(mask, 0.9))

        oval_features = Ellipse(
            radius_x=1,
            radius_y=3,
            center=Point(0, 0),
            rotation=0,
            id="id",
            children=[],
            parents=[],
            label="label",
        )
        expected = {
            "area": 9.42,
            "averageDiameter": 4,
            "center": {"x": 0, "y": 0},
            "children": [],
            "circularity": 0.6,
            "id": "id",
            "label": "label",
            "minorAxis": 2,
            "parents": [],
            "radiusX": 1,
            "radiusY": 3,
            "angleOfRotation": 0,
            "type": "ellipse",
            "accuracy": 1,
        }

        actual = oval_features.to_dict(
            features=["area", "circularity", "minor_axis", "average_diameter"]
        )
        self.assertDictEqual(expected, actual)

    def test_feature_selection_empty(self):
        oval_features = Ellipse(
            radius_x=1,
            radius_y=3,
            center=Point(0, 0),
            rotation=0,
            id="id",
            children=[],
            parents=[],
            label="label",
        )
        expected = {
            "center": {"x": 0, "y": 0},
            "children": [],
            "id": "id",
            "label": "label",
            "parents": [],
            "radiusX": 1,
            "radiusY": 3,
            "angleOfRotation": 0,
            "type": "ellipse",
            "accuracy": 1,
        }

        actual = oval_features.to_dict(features=[])
        self.assertDictEqual(expected, actual)

    def test_feature_selection_None(self):
        oval_features = Ellipse(
            radius_x=1,
            radius_y=3,
            center=Point(0, 0),
            rotation=0,
            id="id",
            children=[],
            parents=[],
            label="label",
        )
        expected = {
            "area": 9.42,
            "aspectRatio": 0.33,
            "averageDiameter": 4,
            "center": {"x": 0, "y": 0},
            "children": [],
            "circularity": 0.6,
            "convexHullArea": 9.42,
            "id": "id",
            "label": "label",
            "majorAxis": 6,
            "minorAxis": 2,
            "parents": [],
            "perimeter": 14.05,
            "radiusX": 1,
            "radiusY": 3,
            "angleOfRotation": 0,
            "type": "ellipse",
            "accuracy": 1,
            "roundness": 0.33,
        }

        actual = oval_features.to_dict(features=None)
        self.assertDictEqual(expected, actual)

    def test_feature_multiplier(self):
        feature_multiplier = 742
        oval_features = Ellipse(
            radius_x=1,
            radius_y=3,
            center=Point(0, 0),
            rotation=0,
            id="id",
            children=[],
            parents=[],
            label="label",
        )
        expected = {
            "area": round(9.42 * feature_multiplier**2, 2),
            "aspectRatio": 0.33,
            "averageDiameter": round(4 * feature_multiplier, 2),
            "center": {"x": 0, "y": 0},
            "children": [],
            "circularity": 0.6,
            "convexHullArea": round(9.42 * feature_multiplier**2, 2),
            "id": "id",
            "label": "label",
            "majorAxis": round(6 * feature_multiplier, 2),
            "minorAxis": round(2 * feature_multiplier, 2),
            "parents": [],
            "perimeter": round(14.05 * feature_multiplier, 2),
            "radiusX": 1,
            "radiusY": 3,
            "angleOfRotation": 0,
            "type": "ellipse",
            "accuracy": 1,
            "roundness": 0.33,
        }

        actual = oval_features.to_dict(feature_multiplier=742)
        assertAlmostEqualDictEqual(self, expected, actual, 1)

    def test_metric_unit(self):
        oval_features = Ellipse(
            radius_x=1,
            radius_y=3,
            center=Point(0, 0),
            rotation=0,
            id="id",
            children=[],
            parents=[],
            label="label",
        )
        expected = {
            "areaPm": 9.42,
            "aspectRatio": 0.33,
            "averageDiameterPm": 4,
            "center": {"x": 0, "y": 0},
            "children": [],
            "circularity": 0.6,
            "convexHullAreaPm": 9.42,
            "id": "id",
            "label": "label",
            "majorAxisPm": 6,
            "minorAxisPm": 2,
            "parents": [],
            "perimeterPm": 14.05,
            "radiusX": 1,
            "radiusY": 3,
            "angleOfRotation": 0,
            "type": "ellipse",
            "accuracy": 1,
            "roundness": 0.33,
        }

        actual = oval_features.to_dict(features=None, metric_unit="pm")
        self.assertDictEqual(expected, actual)



from typing import List

from tomni.annotation_manager.annotations.point import Point


def are_lines_equal(l1: List[Point], l2: List[Point], is_enclosed: bool = True):
    if is_enclosed == False:
        raise NotImplemented()

    if len(l1) != len(l2):
        return False

    # Find the first element in l2 that is the same as l1[0]
    # This is the start of the comparing
    start_l2 = 0
    for i in range(len(l1)):
        if l1[0] == l2[i]:
            start_l2 = i
            break

    # sliced_l2 and l1 have the same starting point
    sliced_l2 = l2[start_l2:] + l2[:start_l2]
    print(sliced_l2)

    for i in range(len(l1)):
        if l1[i] != sliced_l2[i]:
            return False

    return True



# -*- coding: utf-8 -*-
"""Tomni is a collection of image analysis functions useful for CytoSmart solution."""

__author__ = """Tom Nijhof & Jelle van Kerkvoorde & Bram van der Velden"""
__email__ = "bram.vandervelden@axionbio.com & jelle.vankerkvoorde@axionbio.com"
__version__ = "2.2.1b0"

from .bbox_fitting import bbox_fitting, bbox_fitting_center
from .bbox_operations import check_overlap_bbox
from .convert_color import convert_color
from .illumination_correction import (
    absolute_difference,
    fluo_tophat,
    relative_difference,
)
from .img_dim import img_dim
from .img_paste import img_paste
from .json_operations import *
from .make_mask import *
from .shape_fitting import fit_rect_around_ellipse
from .transformers import *



from typing import Union, List

import numpy as np

from tomni.annotation_manager.annotations.point import Point


def parse_points_to_contour(points: List[Point]) -> np.ndarray:
    contour_points = [[[point.x, point.y]] for point in points]
    contour = np.array(contour_points, dtype=np.int32)
    return contour


def parse_points_to_inner_contour(
    list_points: Union[List[List[Point]], List]
) -> List[np.ndarray]:
    inner_contours = []
    for points in list_points:
        contour_points = [[[point.x, point.y]] for point in points]
        contour = np.array(contour_points, dtype=np.int32)
        inner_contours.append(contour)
    return inner_contours



from simplification.cutil import simplify_coords
from typing import List
from tomni.annotation_manager.annotations.point import Point


def compress_polygon_points(
    points: List[Point], epsilon: float = 3.0, min_number_of_points: int = 5
):
    """Compresses the polygon

    Args:
        points (List): The points of the polygon
        epsilon (float, optional): Epsilon value in Ramer-Douglas-Peucker algorithm. The smaller the epsilon, the more points are kept and the more faithful the simplified curve is to the original one.
        Defaults to 3.0.

    Returns:
        _type_: A list of 2D points, each represented as an instance of the `Point` class.
    """
    assert epsilon >= 0

    point_arr = [[point.x, point.y] for point in points]
    point_arr_compressed = simplify_coords(point_arr, epsilon)
    if len(point_arr_compressed) < min_number_of_points:
        return points
    return [Point(x, y) for (x, y) in point_arr_compressed]



from unittest import TestCase

from tomni.annotation_manager.annotations.point import Point
from .main import are_lines_equal


class TestLineEqual(TestCase):
    def test_happy_flow(self):
        line_1 = [Point(1, 1), Point(1, 30), Point(30, 30), Point(30, 1)]
        line_2 = [Point(1, 1), Point(1, 30), Point(30, 30), Point(30, 1)]

        self.assertTrue(are_lines_equal(line_1, line_2))

    def test_rotated(self):
        line_1 = [Point(1, 1), Point(1, 30), Point(30, 30), Point(30, 1)]
        line_2 = [Point(30, 30), Point(30, 1), Point(1, 1), Point(1, 30)]

        self.assertTrue(are_lines_equal(line_1, line_2))

    def test_different(self):
        line_1 = [Point(1, 1), Point(1, 30), Point(30, 30), Point(30, 1)]
        line_2 = [Point(30, 30), Point(30, 1), Point(1, 2), Point(1, 30)]

        self.assertFalse(are_lines_equal(line_1, line_2))

    def test_different_length(self):
        line_1 = [Point(1, 1), Point(1, 30), Point(30, 30), Point(30, 1)]
        line_2 = [Point(30, 30), Point(30, 1), Point(1, 1), Point(1, 1), Point(1, 30)]

        self.assertFalse(are_lines_equal(line_1, line_2))



import uuid
from typing import Dict, List, Tuple, Union

import cv2
import numpy as np

from tomni.annotation_manager.utils.contours2polygons import contours2polygons
from .annotations import Annotation, Ellipse, Point, Polygon
from .utils import parse_points_to_contour

MIN_NR_POINTS_POLYGON = 5


class AnnotationManager(object):
    def __init__(self, annotations: List[Annotation]):
        """Initializes a AnnotationManager object.

        Args:
            annotations (List[Annotation]): Collection of annotations, e.g. polygon or ellipse.
        """
        self._annotations = annotations

    @classmethod
    def from_dicts(
        cls,
        dicts: List[dict],
    ):
        """
        Initializes the class with a list of dictionaries containing annotations.

        Args:
            cls ('AnnotationManager'): The class itself.
            dicts (List[dict]): A list of dicts containing annotations.

        Raises:
            ValueError: Raised if the input dictionaries are not properly formatted.

        Note:
            Only Polygon and Ellipse annotations are supported.

        Returns:
            AnnotationManager: An instance of the AnnotationManager class containing the parsed annotations.
        """
        TYPE_KEY = "type"
        LABEL_KEY = "label"
        CHILDREN_KEY = "children"
        PARENTS_KEY = "parents"
        ID_KEY = "id"
        CENTER_KEY = "center"
        annotations = []

        for d in dicts:
            if d[TYPE_KEY] == "ellipse":
                annotation = Ellipse(
                    label=d.get(LABEL_KEY, None),
                    id=d.get(ID_KEY, str(uuid.uuid4())),
                    children=d.get(CHILDREN_KEY, []),
                    parents=d.get(PARENTS_KEY, []),
                    radius_x=d["radiusX"],
                    radius_y=d.get("radiusY", None),
                    center=Point(x=d[CENTER_KEY]["x"], y=d[CENTER_KEY]["y"]),
                    rotation=d["angleOfRotation"],
                    accuracy=d.get("accuracy", 1),
                )
            elif d[TYPE_KEY] == "polygon":
                if len(d["points"]) < MIN_NR_POINTS_POLYGON:
                    continue
                if "inner_points" in d:
                    inner_points = [
                        [Point(x=pi["x"], y=pi["y"]) for pi in inner_contour]
                        for inner_contour in d["inner_points"]
                    ]
                else:
                    inner_points = []

                annotation = Polygon(
                    label=d.get(LABEL_KEY, None),
                    id=d.get(ID_KEY, str(uuid.uuid4())),
                    children=d.get(CHILDREN_KEY, []),
                    parents=d.get(PARENTS_KEY, []),
                    points=[Point(x=p["x"], y=p["y"]) for p in d["points"]],
                    inner_points=inner_points,
                    accuracy=d.get("accuracy", 1),
                )
            else:
                raise ValueError(
                    f"CDF cannot be created. Dict with id {d.get('id', None)} misses type-key with value ellipse or polygon."
                )
            annotations.append(annotation)

        return cls(annotations)

    @classmethod
    def from_binary_mask(
        cls, mask: np.ndarray, include_inner_contours: bool = False, label: str = ""
    ):
        """
        Initializes an AnnotationManager object from a binary mask.

        Args:
            cls ('AnnotationManager'): The class itself.
            mask (np.ndarray): Binary mask input.
            include_inner_contours (bool, optional): Include annotations that are contained within another annotation.
                Defaults to False.
            label (str, optional): A label to assign to the annotations. Defaults to "".

        Returns:
            AnnotationManager: A new AnnotationManager object created from the binary mask.
        """

        mask = mask.astype(np.uint8)
        mode = cv2.RETR_CCOMP if include_inner_contours else cv2.RETR_EXTERNAL
        contours, hierarchy = cv2.findContours(mask, mode, cv2.CHAIN_APPROX_SIMPLE)

        annotations = contours2polygons(
            contours=contours,
            hierarchy=hierarchy,
            include_inner_contours=include_inner_contours,
            label=label,
        )

        return cls(annotations)

    @classmethod
    def from_labeled_mask(
        cls,
        mask: np.ndarray,
        labels: Union[List[str], str] = "",
        include_inner_contours: bool = False,
    ):
        """
        Initializes an AnnotationManager object from a labeled mask.
        A labeled mask contains components indicated by the same pixel values.

        Args:
            cls ('AnnotationManager'): The class itself.
            mask (np.ndarray): A labeled mask with a maximum number of components limited by max(np.uint32).
            labels (Union[List[str], str], optional): A list of class names to add to Polygon labels. Defaults to "".
                Should have the same number of unique pixel values as classes.
                Class names in order of low pixel value to high pixel value.
            include_inner_contours (bool, optional): Include annotations that are contained within another annotation.
                Defaults to False.

        Example input with multiple pixel values::

            [
                [0, 0, 2, 1, 1],
                [0, 0, 2, 0, 0],
                [0, 0, 2, 0, 0]
            ]

        Returns:
            AnnotationManager: A new AnnotationManager object.
        """
        mode = cv2.RETR_CCOMP if include_inner_contours else cv2.RETR_EXTERNAL

        annotations = []
        if isinstance(labels, str):
            mask = mask.astype(np.uint8)
            _, mask = cv2.threshold(mask, 0, 255, cv2.THRESH_BINARY)
            contours, hierarchy = cv2.findContours(mask, mode, cv2.CHAIN_APPROX_SIMPLE)

            annotations = contours2polygons(
                contours=contours,
                include_inner_contours=include_inner_contours,
                hierarchy=hierarchy,
                label=labels,
            )

        elif isinstance(labels, List):
            unique_values = np.unique(mask)
            unique_values = unique_values[unique_values != 0]

            if len(labels) < len(unique_values):
                raise ValueError(
                    f"Not enough labels for unique pixel values. {len(labels)} labels for {len(unique_values)} unique pixel values."
                )

            # Generate seperate mask for each class and find contours.
            for pixel_value in unique_values:
                class_mask = np.uint8(mask == pixel_value)
                contours, hierarchy = cv2.findContours(
                    class_mask, mode, cv2.CHAIN_APPROX_SIMPLE
                )
                annotations.extend(
                    contours2polygons(
                        contours=contours,
                        hierarchy=hierarchy,
                        include_inner_contours=include_inner_contours,
                        label=labels[pixel_value - 1],
                    )
                )
        else:
            raise ValueError("Labels must be either a string or a list of strings.")

        return cls(annotations)

    @classmethod
    def from_darwin(cls, dicts: List[dict]):
        """must be an option"""
        pass

    @property
    def annotations(self) -> List[Annotation]:
        return self._annotations

    @annotations.setter
    def annotations(self, other_annotations: List[Annotation]):
        """I doubt this setter should be allowed to exist."""

    def __len__(self) -> int:
        return len(self._annotations)

    def __eq__(self, other: object) -> bool:
        """To check equality of to CDF objects
        Thats bit of a tricky one because you must compare all annotations IMO.
        Ex: cdf1 = AnnotationManager.from_something()
        cdf2 = AnnotationManager.from_something()
        is cdf1 == cdf2 must be possible.
        """
        pass

    def __contains__(self, other: Annotation):
        """to check if self.annotations contains other."""
        pass

    def __iter__(self):
        self.idx = 0
        return self

    def __next__(self):
        if self.idx < self.__len__():
            annotation = self.annotations[self.idx]
            self.idx += 1
            return annotation
        else:
            raise StopIteration

    def to_dict(
        self,
        decimals: int = 2,
        mask_json: Union[List[dict], None] = None,
        min_overlap: float = 0.9,
        features: Union[List[str], None] = None,
        metric_unit: str = "",
        feature_multiplier: float = 1,
        **kwargs,
    ) -> List[Dict]:
        """
        Transform the AnnotationManager object into a collection of data in AxionBio format.

        Args:
            decimals (int, optional): The number of decimals to use when rounding. Defaults to 2.
            mask_json (Union[dict, None], optional): The dictionary mask that indicates the area to include in the output dictionary.
                Defaults to None.
            min_overlap (float, optional): Minimum overlap required between the polygon and the mask, expressed as a value between 0 and 1.
                Defaults to 0.9.
            features (Union[List[str], None], optional): The features you want to calculate and add to the dictionary objects.
                Defaults to None, which returns all features.
            metric_unit (str, optional): The suffix to add to the dictionary keys' names in camelCasing. Defaults to "".
            feature_multiplier (float, optional): A multiplier used during feature calculation, e.g., 1/742. Defaults to 1.

        Note:
            - If a `mask_json` is provided, the method filters annotations based on their overlap with the mask.
            - Only annotations meeting the specified `min_overlap` criteria are included in the output.
            - If no `mask_json` is provided, all annotations are included in the output.

        Returns:
            List[Dict]: Output is a list of dictionaries in AxionBio format.
        """

        if mask_json is not None:
            filtered_annotations = self._annotations.copy()
            filtered_annotations = [
                annotation
                for annotation in filtered_annotations
                if annotation.is_in_mask(mask_json, min_overlap)
            ]

            return [
                annotation.to_dict(
                    decimals=decimals,
                    features=features,
                    metric_unit=metric_unit,
                    feature_multiplier=feature_multiplier,
                    **kwargs,
                )
                for annotation in filtered_annotations
            ]

        return [
            annotation.to_dict(
                decimals=decimals,
                features=features,
                metric_unit=metric_unit,
                feature_multiplier=feature_multiplier,
                **kwargs,
            )
            for annotation in self._annotations
        ]

    def to_contours(self) -> List[np.ndarray]:
        """
        Transform an AnnotationManager object to a collection of OpenCV-style contours.

        This method generates a collection of contours from the annotations stored in the AnnotationManager object.
        Supported annotation type for conversion is 'Polygon'. Each contour is represented as a NumPy array of shape
        (N, 1, 2), where N is the number of points in the contour, and each point has (x, y) coordinates.

        Raises:
            ValueError: If any annotation in the AnnotationManager is not of type 'Polygon'.

        Returns:
            List[np.ndarray]: A collection of contours, where each contour is represented as a NumPy array.

        Note:
            - This method supports only annotations of type 'Polygon' for conversion to contours.
            - Each contour is represented as a NumPy array of shape (N, 1, 2), where N is the number of points in
              the contour, and each point has (x, y) coordinates.
        """

        if not all(
            [isinstance(annotation, Polygon) for annotation in self._annotations]
        ):
            raise ValueError("`to_contours is only supported on polygon-annotations.`")

        contours = [
            parse_points_to_contour(annotation.points)
            for annotation in self._annotations
        ]

        return contours

    def to_binary_mask(self, shape: Tuple[int, int]) -> np.ndarray:
        """
        Transform an AnnotationManager object to a binary mask.

        This method generates a binary mask from the annotations stored in the AnnotationManager object.
        Supported annotation types for conversion are polygon and ellipse.

        Args:
            shape (Tuple[int, int]): The shape (width, height) of the new binary mask.

        Returns:
            np.ndarray: A binary mask where annotated regions are represented by 1 (True) and
            non-annotated regions are represented by 0 (False).

        Note:
            - This method supports annotations of type Polygon and Ellipse for conversion to a binary mask.
            - The binary mask represents annotated regions with 1 and non-annotated regions with 0.
        """

        mask = np.zeros(shape, dtype=np.uint8)
        for annotation in self.annotations:
            mask = cv2.bitwise_or(mask, annotation.to_binary_mask(shape))

        return mask

    def to_labeled_mask(self, shape: Tuple[int, int]) -> np.ndarray:
        """
        Transform an Annotation Manager object to a labeled mask. This method generates a labeled mask from the annotations stored in the AnnotationManager object. Supported
        annotation types for conversion are polygon and ellipse.

        Args:
            shape (Tuple[int, int]): The shape (width, height) of the new labeled mask.

        Returns:
            np.ndarray: A new labeled mask where each labeled region corresponds to an annotation.

        Raises:
            TypeError: If the AnnotationManager contains annotations of unsupported types.

        Note:
            - This method supports annotations of type Polygon and Ellipse for conversion to a labeled mask.
            - Each labeled region in the generated mask corresponds to an annotation, and the regions are labeled with unique integer values starting from 1.
        """

        mask = np.zeros(shape, dtype=np.uint8)
        label_color = 1

        for annotation in self._annotations:
            if isinstance(annotation, Polygon):
                points = np.array(
                    [[point.x, point.y] for point in annotation.points], dtype=np.int32
                )
                cv2.fillPoly(mask, [points], color=label_color)
            elif isinstance(annotation, Ellipse):
                cv2.ellipse(
                    mask,
                    center=(annotation.center.x, annotation.center.y),
                    axes=(annotation.radius_x, annotation.radius_y),
                    angle=annotation.rotation,
                    startAngle=0,
                    endAngle=360,
                    color=label_color,
                    thickness=-1,
                )
            else:
                raise TypeError(
                    "Innapropiate annotation type for `to_labeled_mask`. Supported annotations are ellipse and polygon."
                )

            # increase color for every annotation.
            label_color += 1

        return mask

    def to_darwin(self) -> List[Dict]:
        """
        TODO: Convert annotations to darwin format (v7).
        """

    def __add__(self, other):
        """Ability to add to CDF objects together
        cdf1 + cdf2.

        or possiby
        - cdf + dict
        - cdf + darwin
        - ...
        """
        pass

    def __radd__(self, other):
        """Reverse of __add__
        if you do:
        dict + cdf -> error
        radd should flip the two parts and call add.
        so, dict + cdf becomes cdf + dict.
        """
        pass

    def delete_annotation(self, item: Annotation):
        """Remove an annotation from self.annotations.
        Find and delete.
        """
        pass

    def filter(
        self,
        feature: str,
        min_val: float,
        max_val: float,
        feature_multiplier: float = 1.0,
        inplace: bool = False,
    ):
        """
        Filter annotations based on a specified feature.

        This method filters annotations in the AnnotationManager based on a given feature within a specified value range.
        Annotations that meet the filtering criteria are included in the result.

        Args:
            feature (str): The name of the feature to use for filtering, e.g., 'roundness' or 'area'.
            min_val (float): The minimum threshold value for the feature.
            max_val (float): The maximum threshold value for the feature.
            feature_multiplier (float, optional): A multiplier used in feature calculations. Defaults to 1.
            inplace (bool, optional): If True, filter annotations in-place, modifying the object internally.
                If False, return a collection of filtered annotations without modifying the original object. Defaults to False.

        Returns:
            Union[AnnotationManager, List[Annotation]]: If `inplace=True`, returns the AnnotationManager object
            with the filtered annotations. If `inplace=False`, returns a list of filtered annotations.

        Note:
            - This method filters annotations based on a specified feature within the provided value range.
            - The `feature_multiplier` parameter allows scaling of the feature calculation if needed.
        """

        filtered_annotations = []

        for annotation in self._annotations:
            annotation._feature_multiplier = feature_multiplier
            feature_value = getattr(annotation, feature)
            if min_val <= feature_value <= max_val:
                filtered_annotations.append(annotation)

        if inplace:
            self._annotations = filtered_annotations
            return self

        return filtered_annotations

    @classmethod
    def get_circularity_summary(self):
        """loops the cdf items to get avg, std, min, max."""

    def get_feature_summaries(self, features: List[str]) -> Dict:
        """Pass a list of features to calculated.

        This is not ideal and i am for suggestion. I want to avoid having to calculate over and over.
        """
        circularity = []
        for cdf_item in self._cdf_data:
            circularity.append(cdf_item.circularity)

        return {
            "circularity": {"avg": 1, "std": 1, "min": 1, "max": 1},
            "...": "...",
            "feature_n": {"avg_n": 1, "std_n": 1, "min_n": 1, "max_n": 1},
        }



import numpy as np


def mask2bbox(mask: np.ndarray, padding: int = 0) -> tuple:
    """
    Convert a binary image mask into bounding box coordinates (xmin, ymin, xmax, ymax).

    Args:
        mask (np.ndarray): A binary image mask represented as a NumPy array where 1 indicates the object
            of interest, and 0 indicates the background.
        padding (int, optional): Padding added or subtracted from the bounding box coordinates.
            The resulting bounding box coordinates will not exceed the image dimensions. If padding is positive,
            it adds padding to all sides of the bounding box; if negative, it subtracts padding.
            Defaults to 0.

    Returns:
        tuple: A tuple containing the bounding box coordinates (xmin, ymin, xmax, ymax).
            The coordinates represent the top-left and bottom-right corners of the bounding box.
    """
    x1, x2, y1, y2 = (0, 0, 0, 0)

    max_x = np.max(mask, axis=0)
    max_y = np.max(mask, axis=1)

    x_nonzeros = np.nonzero(max_x)
    y_nonzeros = np.nonzero(max_y)

    if x_nonzeros:
        if len(x_nonzeros[0] > 0):
            x1 = np.nonzero(max_x)[0][0]
            x2 = np.nonzero(max_x)[0][-1] + 1
    else:
        return (x1, y1, x2, y2)

    if y_nonzeros:
        if len(y_nonzeros[0]) > 0:
            y1 = np.nonzero(max_y)[0][0]
            y2 = np.nonzero(max_y)[0][-1] + 1
    else:
        return (x1, y1, x2, y2)

    if x1 == y1 == x2 == y2 == 0:
        return (x1, y1, x2, y2)

    # Add padding to bounding box
    if x1 - padding < 0:
        x1 = 0
    else:
        x1 -= padding

    if x2 + padding >= mask.shape[1]:
        x2 = mask.shape[1]
    else:
        x2 += padding

    if y1 - padding < 0:
        y1 = 0
    else:
        y1 -= padding

    if y2 + padding >= mask.shape[0]:
        y2 = mask.shape[0]
    else:
        y2 += padding

    return (x1, y1, x2, y2)



import numpy as np
import cv2
import json

from tomni.transformers import json2vgg

# README
# To test in VIA, first import the json file and than add the image

img_name = "test_vgg"
extension = ".jpg"
img = np.zeros((100, 100))

json_list = [
    {
        "type": "polygon",
        "accuracy": 0.4,
        "parents": [],
        "children": [],
        "id": "unicorn",
        "points": [{"x": 50, "y": 74}, {"x": 47, "y": 75}, {"x": 44, "y": 89}],
    },
    {
        "type": "ellipse",
        "accuracy": 0.32,
        "id": "unicorn",
        "center": {"x": 76, "y": 64},
        "radiusX": 5,
        "radiusY": 5,
        "angleOfRotation": 0,
    },
]

vgg_json = json2vgg(json_list, img_name, extension)

cv2.imwrite(f"{img_name}{extension}", img)

with open(f"{img_name}.json", "w") as f:
    json.dump(vgg_json, f)



from dataclasses import asdict
from unittest import TestCase

from tomni.annotation_manager import Ellipse, Point, Polygon
from tomni.annotation_manager.main import AnnotationManager


class TestPolygon(TestCase):
    def setUp(self) -> None:
        id_ = "1234-1234-2134-1321"
        children = []
        parents = []
        label = "polygon_test"

        self.maxDiff = None
        self.circular_points = [
            Point(1, 2),
            Point(1, 4),
            Point(2, 5),
            Point(4, 5),
            Point(5, 4),
            Point(5, 2),
            Point(4, 1),
            Point(3, 1),
            Point(2, 1),
        ]
        self.circular_polygon = Polygon(
            points=self.circular_points,
            id=id_,
            children=children,
            parents=parents,
            label=label,
        )

        self.star_shaped_points = [
            Point(1, 3),
            Point(2, 3),
            Point(3, 5),
            Point(5, 3),
            Point(3, 1),
            Point(2, 2),
        ]
        self.star_shaped_polygon = Polygon(
            points=self.star_shaped_points,
            id=id_,
            children=children,
            parents=parents,
            label=label,
        )

        self.rectangle_points = [
            Point(1, 5),
            Point(3, 5),
            Point(5, 5),
            Point(5, 1),
            Point(3, 1),
            Point(1, 1),
        ]
        self.rectangle_polygon = Polygon(
            points=self.rectangle_points,
            id=id_,
            children=children,
            parents=parents,
            label=label,
        )

    # Area
    def test_circular_area(self):
        expected = 14.0
        actual = self.circular_polygon.area

        self.assertEqual(expected, actual)

    def test_rectangular_area(self):
        expected = 16.0
        actual = self.rectangle_polygon.area

        self.assertEqual(expected, actual)

    def test_star_shaped_area(self):
        expected = 7.0
        actual = self.star_shaped_polygon.area

        self.assertEqual(expected, actual)

    # Circularity
    def test_circular_circularity(self):
        expected = 0.9432711611657616
        actual = self.circular_polygon.circularity

        self.assertEqual(expected, actual)

    def test_star_shaped_circularity(self):
        expected = 0.640254577329347
        actual = self.star_shaped_polygon.circularity

        self.assertEqual(expected, actual)

    def test_rectangular_circularity(self):
        expected = 0.7853981633974483
        actual = self.rectangle_polygon.circularity

        self.assertEqual(expected, actual)

    # Convex_hull_area
    def test_circular_convex_hull_area(self):
        expected = 14.0
        actual = self.circular_polygon.convex_hull_area

        self.assertEqual(expected, actual)

    def test_star_shaped_convex_hull_area(self):
        expected = 8.0
        actual = self.star_shaped_polygon.convex_hull_area

        self.assertEqual(expected, actual)

    def test_rectangular_convex_hull_area(self):
        expected = 16.0
        actual = self.rectangle_polygon.convex_hull_area

        self.assertEqual(expected, actual)

    # Perimeter
    def test_circular_perimeter(self):
        expected = 13.656854152679443
        actual = self.circular_polygon.perimeter

        self.assertEqual(expected, actual)

    def test_star_shaped_perimeter(self):
        expected = 11.721349239349365
        actual = self.star_shaped_polygon.perimeter

        self.assertEqual(expected, actual)

    def test_rectangular_perimeter(self):
        expected = 16.0
        actual = self.rectangle_polygon.perimeter

        self.assertEqual(expected, actual)

    # Minor Axis
    def test_circular_minor_axis(self):
        expected = 4.363824844360352
        actual = self.circular_polygon.minor_axis

        self.assertAlmostEqual(expected, actual)

    def test_rectangular_minor_axis(self):
        expected = 5.551115205843844e-18
        actual = self.rectangle_polygon.minor_axis

        self.assertAlmostEqual(expected, actual)

    def test_star_shaped_minor_axis(self):
        expected = 2.985952854156494
        actual = self.star_shaped_polygon.minor_axis

        self.assertAlmostEqual(expected, actual)

    # Major Axis
    def test_circular_major_axis(self):
        expected = 4.526668071746826
        actual = self.circular_polygon.major_axis

        self.assertEqual(expected, actual)

    def test_rectangular_major_axis(self):
        expected = 4.0
        actual = self.rectangle_polygon.major_axis

        self.assertEqual(expected, actual)

    def test_star_shaped_major_axis(self):
        expected = 7.537467002868652
        actual = self.star_shaped_polygon.major_axis

        self.assertEqual(expected, actual)

    # Average Diameter
    def test_circular_average_diameter(self):
        expected = 4.445246458053589
        actual = self.circular_polygon.average_diameter

        self.assertEqual(expected, actual)

    def test_rectangular_average_diameter(self):
        expected = 2.0
        actual = self.rectangle_polygon.average_diameter

        self.assertEqual(expected, actual)

    def test_star_shaped_average_diameter(self):
        expected = 5.261709928512573
        actual = self.star_shaped_polygon.average_diameter

        self.assertEqual(expected, actual)

    # Aspect ratio
    def test_circular_aspect_ratio(self):
        expected = 0.9640258077673378
        actual = self.circular_polygon.aspect_ratio

        self.assertEqual(expected, actual)

    def test_rectangular_aspect_ratio(self):
        expected = 1.387778801460961e-18
        actual = self.rectangle_polygon.aspect_ratio

        self.assertEqual(expected, actual)

    def test_star_shaped_aspect_ratio(self):
        expected = 0.39614804987140684
        actual = self.star_shaped_polygon.aspect_ratio

        self.assertEqual(expected, actual)

    # Roundness
    def test_circular_roundness(self):
        expected = 0.8911880248803259
        actual = self.circular_polygon.roundness

        self.assertEqual(expected, actual)

    def test_rectangular_roundness(self):
        expected = 0.6365748269154868
        actual = self.rectangle_polygon.roundness

        self.assertEqual(expected, actual)

    def test_star_shaped_roundness(self):
        expected = 0.5569866579216156
        actual = self.star_shaped_polygon.roundness

        self.assertEqual(expected, actual)

    def test_circular_to_dict(self):
        expected = {
            "id": "1234-1234-2134-1321",
            "label": "polygon_test",
            "children": [],
            "parents": [],
            "type": "polygon",
            "area": 14.0,
            "convexHullArea": 14.0,
            "majorAxis": 4.53,
            "minorAxis": 4.36,
            "averageDiameter": 4.45,
            "aspectRatio": 0.96,
            "circularity": 0.94,
            "perimeter": 13.66,
            "roundness": 0.89,
            "points": [asdict(point) for point in self.circular_points],
            "accuracy": 1,
            "inner_points": [],
        }
        actual = self.circular_polygon.to_dict()

        self.assertDictEqual(expected, actual)

    def test_rectangular_to_dict(self):
        expected = {
            "id": "1234-1234-2134-1321",
            "label": "polygon_test",
            "children": [],
            "parents": [],
            "type": "polygon",
            "area": 16.0,
            "majorAxis": 4.0,
            "minorAxis": 0.0,
            "averageDiameter": 2.0,
            "aspectRatio": 0.0,
            "circularity": 0.79,
            "convexHullArea": 16.0,
            "perimeter": 16.0,
            "roundness": 0.64,
            "points": [asdict(point) for point in self.rectangle_points],
            "accuracy": 1,
            "inner_points": [],
        }
        actual = self.rectangle_polygon.to_dict()

        self.assertDictEqual(expected, actual)

    def test_star_shaped_to_dict(self):
        expected = {
            "id": "1234-1234-2134-1321",
            "label": "polygon_test",
            "children": [],
            "parents": [],
            "type": "polygon",
            "area": 7.0,
            "minorAxis": 2.99,
            "majorAxis": 7.54,
            "averageDiameter": 5.26,
            "aspectRatio": 0.40,
            "circularity": 0.64,
            "convexHullArea": 8.0,
            "perimeter": 11.72,
            "roundness": 0.56,
            "points": [asdict(point) for point in self.star_shaped_points],
            "accuracy": 1,
            "inner_points": [],
        }
        actual = self.star_shaped_polygon.to_dict()

        self.assertDictEqual(expected, actual)

    def test_triangle_to_dict(self):
        self.triangle_points = [Point(1, 5), Point(3, 1), Point(5, 5)]
        with self.assertRaises(ValueError):
            Polygon(
                points=self.triangle_points,
                id="id",
                children=[],
                parents=[],
                label="label",
            )

    def test_accuracy_to_dict(self):
        expected = {
            "id": "1234-1234-2134-1321",
            "label": "polygon_test",
            "children": [],
            "parents": [],
            "type": "polygon",
            "area": 14.0,
            "convexHullArea": 14.0,
            "majorAxis": 4.53,
            "minorAxis": 4.36,
            "averageDiameter": 4.45,
            "aspectRatio": 0.96,
            "circularity": 0.94,
            "perimeter": 13.66,
            "roundness": 0.89,
            "points": [asdict(point) for point in self.circular_points],
            "accuracy": 0.5,
            "inner_points": [],
        }

        accuracy_test_object = Polygon(
            points=self.circular_points,
            id="1234-1234-2134-1321",
            children=[],
            parents=[],
            label="polygon_test",
            accuracy=0.5,
        )

        actual = accuracy_test_object.to_dict()
        self.assertDictEqual(expected, actual)

    def test_accuracy_to_dict2(self):
        expected = {
            "id": "1234-1234-2134-1321",
            "label": "polygon_test",
            "children": [],
            "parents": [],
            "type": "polygon",
            "area": 14.0,
            "convexHullArea": 14.0,
            "majorAxis": 4.53,
            "minorAxis": 4.36,
            "averageDiameter": 4.45,
            "aspectRatio": 0.96,
            "circularity": 0.94,
            "perimeter": 13.66,
            "roundness": 0.89,
            "points": [asdict(point) for point in self.circular_points],
            "accuracy": 0,
            "inner_points": [],
        }

        accuracy_test_object = Polygon(
            points=self.circular_points,
            id="1234-1234-2134-1321",
            children=[],
            parents=[],
            label="polygon_test",
            accuracy=0,
        )

        actual = accuracy_test_object.to_dict()
        self.assertDictEqual(expected, actual)

    def test_feasture_multiplier_to_dict(self):
        expected = {
            "id": "1234-1234-2134-1321",
            "label": "polygon_test",
            "children": [],
            "parents": [],
            "type": "polygon",
            "area": 56.0,
            "convexHullArea": 56.0,
            "majorAxis": 9.05,
            "minorAxis": 8.73,
            "averageDiameter": 8.89,
            "aspectRatio": 0.96,
            "circularity": 0.94,
            "perimeter": 27.31,
            "roundness": 0.89,
            "points": [asdict(point) for point in self.circular_points],
            "accuracy": 0.5,
            "inner_points": [],
        }

        accuracy_test_object = Polygon(
            points=self.circular_points,
            id="1234-1234-2134-1321",
            children=[],
            parents=[],
            label="polygon_test",
            accuracy=0.5,
        )

        actual = accuracy_test_object.to_dict(feature_multiplier=2)
        self.assertDictEqual(expected, actual)

    def test_to_dict_with_ellipse_mask(self):
        center = int(2072 / 2)
        rad = int(2072 / 3)

        mask = AnnotationManager(
            [
                Ellipse(
                    center=Point(center, center),
                    radius_x=rad,
                    rotation=0,
                    id="",
                    label="",
                    children=[],
                    parents=[],
                )
            ]
        ).to_dict()

        polygon1 = Polygon(
            points=[
                Point(center, center),
                Point(center + 200, center + 200),
                Point(center, center + 200),
                Point(center - 199, center + 200),
                Point(center - 200, center + 200),
            ],
            id="",
            label="",
            children=[],
            parents=[],
        )
        polygon2 = Polygon(
            points=[
                Point(center, center),
                Point(center - 300, center - 200),
                Point(center, center + 198),
                Point(center, center + 199),
                Point(center, center + 200),
            ],
            id="",
            label="",
            children=[],
            parents=[],
        )
        polygon3 = Polygon(
            points=[
                Point(center, center),
                Point(center + 500, center + 500),
                Point(center + 198, center),
                Point(center + 199, center),
                Point(center + 200, center),
            ],
            id="",
            label="",
            children=[],
            parents=[],
        )

        polygons_inside = [polygon1, polygon2, polygon3]
        polygons_outside = [
            self.star_shaped_polygon,
            self.rectangle_polygon,
        ]

        for polygon in polygons_inside:
            self.assertTrue(polygon.is_in_mask(mask, 0.9))

        for polygon in polygons_outside:
            self.assertFalse(polygon.is_in_mask(mask, 0.9))

    def test_to_dict_with_polygon_mask(self):
        center = 2072 / 2
        size = 2072
        quadrant = size / 4
        points = [
            Point(quadrant, quadrant),
            Point(quadrant, quadrant * 3),
            Point(quadrant * 3, quadrant * 3),
            Point(quadrant * 3, quadrant),
            Point(quadrant * 3, quadrant + 1),
        ]
        mask = AnnotationManager(
            [Polygon(points, id="", label="", children=[], parents=[])]
        ).to_dict()

        polygon1 = Polygon(
            points=[
                Point(center, center),
                Point(center + 200, center + 200),
                Point(center, center + 198),
                Point(center, center + 199),
                Point(center, center + 200),
            ],
            id="",
            label="",
            children=[],
            parents=[],
        )
        polygon2 = Polygon(
            points=[
                Point(center, center),
                Point(center - 300, center - 200),
                Point(center, center + 198),
                Point(center, center + 199),
                Point(center, center + 200),
            ],
            id="",
            label="",
            children=[],
            parents=[],
        )
        polygon3 = Polygon(
            points=[
                Point(center, center),
                Point(center + 500, center + 500),
                Point(center + 200, 198),
                Point(center + 200, 199),
                Point(center + 200, center),
            ],
            id="",
            label="",
            children=[],
            parents=[],
        )

        polygons_inside = [polygon1, polygon2, polygon3]
        polygons_outside = [
            self.star_shaped_polygon,
            self.rectangle_polygon,
        ]

        for polygon in polygons_inside:
            self.assertTrue(polygon.is_in_mask(mask, 0.9))

        for polygon in polygons_outside:
            self.assertFalse(polygon.is_in_mask(mask, 0.9))

    def test_raises(self):
        with self.assertRaises(SyntaxError):
            self.rectangle_polygon.points = self.circular_points

    def test_inner_points(self):
        id = "1234-1234-2134-1321"
        children = []
        parents = []
        label = "inner_point_test"

        self.maxDiff = None
        points = [
            Point(0, 0),
            Point(0, 100),
            Point(0, 200),
            Point(0, 300),
            Point(0, 400),
            Point(100, 400),
            Point(200, 400),
            Point(300, 400),
            Point(400, 400),
            Point(400, 0),
        ]
        inner_points = [
            [
                Point(50, 50),
                Point(50, 100),
                Point(50, 150),
                Point(100, 150),
                Point(150, 150),
                Point(150, 50),
            ]
        ]

        expected = {
            "accuracy": 1,
            "label": label,
            "id": id,
            "children": children,
            "parents": parents,
            "type": "polygon",
            "area": 150000.0,
            "points": [asdict(point) for point in points],
            "inner_points": [
                [asdict(point) for point in list_of_points]
                for list_of_points in inner_points
            ],
        }

        donut_polygon = Polygon(
            points=points,
            id=id,
            children=children,
            parents=parents,
            label=label,
            inner_points=inner_points,
        )
        actual = donut_polygon.to_dict(features=["area"])
        self.assertDictEqual(expected, actual)



import cv2
import numpy as np


def contour2bbox(contour: np.ndarray) -> tuple:
    """
    Convert OpenCV contours into a bounding box format (xmin, ymin, xmax, ymax).

    Args:
        contour (np.ndarray): An array with coordinates defining the contour of an object.

    Returns:
        tuple: A tuple containing the bounding box coordinates (xmin, ymin, xmax, ymax).

    Example::

        contour = np.array([[[3, 3]], [[3, 5]], [[5, 5]], [[5, 3]]], dtype=np.int32)
        bbox = contour2bbox(contour)
        print(bbox)
        (3, 3, 6, 6)

    Note:
        - The function uses OpenCV's `cv2.boundingRect` to compute the bounding box.

    """
    x_min, y_min, w, h = cv2.boundingRect(contour)
    return (x_min, y_min, x_min + w, y_min + h)



from unittest import TestCase
import numpy as np
from .main import ellipse2json


class TestEllipse2json(TestCase):
    def test_happy_flow_circle(self):
        expectedResult = {
            "type": "ellipse",
            "center": {"x": 20, "y": 30},
            "radiusX": 10,
            "radiusY": 10,
            "angleOfRotation": 0,
            "id": "unicorn",
        }

        result = ellipse2json(20, 30, 10)
        result["id"] = "unicorn"

        self.assertDictEqual(result, expectedResult)

    def test_happy_flow_ellipse(self):
        expectedResult = {
            "type": "ellipse",
            "center": {"x": 20, "y": 30},
            "radiusX": 10,
            "radiusY": 15,
            "angleOfRotation": 0.98,
            "id": "unicorn",
        }

        result = ellipse2json(20, 30, 10, 15, 0.98)
        result["id"] = "unicorn"

        self.assertDictEqual(result, expectedResult)



import cv2
import numpy as np


def circularity(contour: np.ndarray) -> float:
    """
    Calculate the circularity of a contour.

    The circularity of a contour is a measure of how closely it resembles a perfect circle.
    It is calculated based on the contour's area and perimeter.

    Args:
        contour (np.ndarray): The contour represented as an OpenCV ndarray.

    Returns:
        float: The circularity value, where 0 < circularity <= 1.

    Example:
        >>> contour = np.array([[[1, 2]], [[2, 3]], [[3, 2]], [[2, 1]]])
        >>> circularity(contour)
        0.7853981633974483  # Approximate circularity of a square.

    Note:
        - Formula: Circularity = (4 * π * area) / (perimeter^2)
        - Circularities close to 0 indicate more irregular shapes, while circularities close to 1 suggest shapes that closely resemble a perfect circle.
    """

    # calculate size
    area = cv2.contourArea(contour)

    # calculate shape
    perimeter = cv2.arcLength(contour, True)

    circularity = (4 * np.pi * area) / (perimeter**2)

    return circularity



import numpy as np
import cv2
from typing import Union


def sendErrorShape(shapeC):
    raise TypeError("Expect contour in form (N, 2) not {}".format(shapeC))


def make_mask_contour(img_shape: tuple, contour: Union[list, np.ndarray]) -> np.ndarray:
    """
    Create a binary mask from a given contour and image shape. The mask sets pixels inside the contour to 1 (True)
    and pixels outside to 0 (False).

    Args:
        img_shape (tuple): The shape of the image given in image coordinates, represented as (width, height).
        contour (list or np.ndarray): The contour as outputted by OpenCV, with shape (N, 2).

    Returns:
        np.ndarray: A boolean image where the area enclosed by the contour is True and the rest is False.

    Raises:
        TypeError: If the provided contour is not in the form (N, 2), where N is the number of points.

    Example::

        # Usage example with OpenCV's contour output
        img_shape = (640, 480)  # Width and height of the target image
        contour = np.array([[100, 200], [200, 100], [300, 400]], dtype=np.int32)
        mask = make_mask_contour(img_shape, contour)
        # 'mask' will contain a binary mask representing the specified contour.
    """

    contour = np.array(contour, dtype=np.int32)
    shapeC = np.shape(contour)

    if len(shapeC) != 2:
        sendErrorShape(shapeC)

    if shapeC[1] != 2:
        sendErrorShape(shapeC)

    contour = contour.reshape((1, shapeC[0], 2))

    mask = np.zeros((img_shape[1], img_shape[0]), dtype=np.uint8)
    cv2.fillPoly(mask, contour, 1)
    return mask



#%%
from tomni.annotation_manager.annotations import Ellipse, Point, Polygon

#%%
# Ellipse annotation
elli = Ellipse(
    radius_x=1,
    radius_y=3,
    center=Point(0, 0),
    rotation=0,
    id="Hello Jan",
    label="cell",
    children=[],
    parents=[],
)

print(elli.to_dict())


#%%
# Polygon annotation
star_shaped_points = [
    Point(1, 3),
    Point(2, 3),
    Point(3, 5),
    Point(5, 3),
    Point(3, 1),
    Point(2, 2),
]
star_shaped_polygon = Polygon(
    points=star_shaped_points,
    id="132132132123132",
    children=[],
    parents=[],
    label="star",
)

print(star_shaped_polygon.to_dict())

# %%



from .contour2bbox import contour2bbox
from .contours2json import contours2json
from .ellipse2json import ellipse2json
from .json2bbox import json2bbox
from .json2contours import json2contours
from .json2dict import json2dict
from .json2labels import json2labels
from .json2mask import json2mask
from .json2vgg import json2vgg
from .labels2contours import labels2contours
from .binary2contours import binary2contours
from .labels2listsOfPoints import labels2listsOfPoints
from .mask2bbox import mask2bbox
from .mask2json import mask2json
from .positions2contour import positions2contour
from .vgg2json import vgg2json



def iterative_downsampling(points: list, n_iter: int) -> list:
    """
    Compress a list of 2D points by reducing the number of points by half iteratively.

    Args:
        points (list): A list of 2D points, where each point is represented as a dictionary with {"x": 1, "y": 0}.
        n_iter (int): An integer representing the number of times to apply the recursive compression method.

    Returns:
        list: A list of 2D points, where each point is represented as a dictionary with {"x": 1, "y": 0}.
            The number of points in the output list is half the number of points in the input list, repeated `n_iter` times.

    """

    assert n_iter >= 0, "Nr of iterations must be zero or positive."

    for _ in range(n_iter):
        points = points[::2]

    return points



from abc import ABC, abstractmethod
from typing import List


class Annotation(ABC):
    def __init__(
        self, id: str, label: str, children: List, parents: List, accuracy: float
    ) -> None:
        self._id: str = id
        self._label: str = label
        self._children: List[Annotation] = children
        self._parents: List[Annotation] = parents
        self._accuracy: float = accuracy

    @abstractmethod
    def to_dict(self, decimals: int = 2, **kwargs) -> dict:
        """Creates a dictionary in CDF of annotation.
        Args:
            decimals (int, optional): The number of decimals to use when rounding. Defaults to 2.

        Returns:
            dict: AxionBio Data Format dict.
        """
        return {
            "id": self._id,
            "label": self._label,
            "children": self._children,
            "parents": self._parents,
            "accuracy": self._accuracy,
        }

    @property
    @abstractmethod
    def label(self) -> str:
        return self._label

    @label.setter
    @abstractmethod
    def label(self, value) -> None:
        self._label = value



import copy
from typing import Union


def rotate_json(json_object: dict, angle: int, img_shape: Union[list, tuple]) -> dict:
    """
    Rotate a JSON object (ellipse or polygon) by 0, 90, 180, or 270 degrees.

    Args:
        json_object (dict): JSON object following the standard AxionBio format (ellipse or polygon).
        angle (int): The angle of rotation. Choose from 0, 90, 180, or 270 degrees.
        img_shape (list or tuple): The shape of the image related to the JSON object, represented as (width, height).

    Returns:
        dict: A new JSON object with the specified rotation.

    Raises:
        ValueError: If the angle is not one of 0, 90, 180, or 270 degrees.
    """

    assert len(img_shape) == 2

    newjson_object = copy.deepcopy(json_object)

    if json_object["type"] == "ellipse":
        if angle != 0:
            newjson_object["center"] = rotate_point(
                json_object["center"], angle, img_shape
            )

    elif json_object["type"] == "polygon":
        if angle != 0:
            newjson_object["points"] = [
                rotate_point(point, angle, img_shape) for point in json_object["points"]
            ]
            if "inner_points" in json_object:
                newjson_object["inner_points"] = [
                    [
                        rotate_point(inner_point, angle, img_shape)
                        for inner_point in inner_contour
                    ]
                    for inner_contour in json_object["inner_points"]
                ]

    else:
        raise ValueError(
            "Unknown JSON object type. Supported types are 'ellipse' and 'polygon'."
        )

    return newjson_object


def rotate_point(point: dict, angle_deg: int, img_shape: Union[list, tuple]) -> dict:
    """Rotate a 2D point by a specified angle in degrees.

    Args:
        point (dict): A dictionary representing the 2D point with 'x' and 'y' coordinates.
        angle_deg (int): The angle of rotation in degrees. Valid values are 0, 90, 180, or 270 degrees.
        img_shape (Union[list, tuple]): The shape of the image related to the JSON object, represented as (width, height).

    Raises:
        ValueError: If the angle_deg is not one of 0, 90, 180, or 270 degrees.

    Returns:
        dict: A new dictionary representing the rotated 2D point.
    """
    x, y = point["x"], point["y"]
    if angle_deg == 90:
        return {"x": img_shape[0] - y - 1, "y": x}
    elif angle_deg == 180:
        return {"x": img_shape[1] - x - 1, "y": img_shape[0] - y - 1}
    elif angle_deg == 270:
        return {"x": y, "y": img_shape[1] - x - 1}
    else:
        raise ValueError("Unknown angle. Angle should be 0, 90, 180, or 270 degrees.")



import math

import cv2

from ...contour_operations import circularity
from ...transformers import json2contours


def add_circularity(json_object: dict) -> None:
    """
    Calculate and add the circularity property to a JSON annotation object.
    The circularity is calculated by drawing a circle around the object and comparing
    the area of the circle with the area of the object.

    Args:
        json_object (dict): A JSON object of type "polygon" following the standard AxionBio format.

    Raises:
        ValueError: If the type of the annotation in the JSON object is not "polygon."

    Note:
        - Only polygon objects are supported for circularity calculation.
        - The circularity value is added as the "circularity" property to the JSON object.
        - The closer the circularity is to 1.0, the more circular the object is.
    """

    if json_object["type"] == "polygon":
        cnt = json2contours(json_object)

        _circularity = circularity(cnt)

        # Add results to current object
        json_object["circularity"] = _circularity
    else:
        raise ValueError(
            f"Currently only the polygon objects are supported; not {json_object['type']}"
        )



from __future__ import absolute_import
from unittest import TestCase
import math
from .main import summary_json


class TestSummaryJson(TestCase):
    def test_summary(self):
        input_json_list = [
            {
                "type": "polygon",
                "accuracy": 0.45455,
                "parents": [],
                "children": [],
                "id": "unicorn",
                "points": [
                    {"x": 0, "y": 10},
                    {"x": 10, "y": 10},
                    {"x": 10, "y": 0},
                    {"x": 0, "y": 0},
                ],
            },
            {
                "type": "ellipse",
                "accuracy": 0.443,
                "id": "unicorn",
                "center": {"x": 769, "y": 649},
                "radiusX": 10,
                "radiusY": 20,
                "angleOfRotation": 0,
            },
        ]

        copy_json_list = [
            {
                "type": "polygon",
                "accuracy": 0.45455,
                "parents": [],
                "children": [],
                "id": "unicorn",
                "points": [
                    {"x": 0, "y": 10},
                    {"x": 10, "y": 10},
                    {"x": 10, "y": 0},
                    {"x": 0, "y": 0},
                ],
            },
            {
                "type": "ellipse",
                "accuracy": 0.443,
                "id": "unicorn",
                "center": {"x": 769, "y": 649},
                "radiusX": 10,
                "radiusY": 20,
                "angleOfRotation": 0,
            },
        ]

        expected_output = (
            2,
            round(math.pi * 10 * 20 + 100, 2),
            round((math.pi * 10 * 20 + 100) / 2, 2),
            round(math.pi * 10 * 20, 2),
            100.0,
        )

        result = summary_json(input_json_list, "area")

        self.assertTupleEqual(result, expected_output)

        self.assertListEqual(input_json_list, copy_json_list)

    def test_summary_copy_false(self):
        input_json_list = [
            {
                "type": "polygon",
                "accuracy": 0.45455,
                "parents": [],
                "children": [],
                "id": "unicorn",
                "points": [
                    {"x": 0, "y": 10},
                    {"x": 10, "y": 10},
                    {"x": 10, "y": 0},
                    {"x": 0, "y": 0},
                ],
            },
            {
                "type": "ellipse",
                "accuracy": 0.443,
                "id": "unicorn",
                "center": {"x": 769, "y": 649},
                "radiusX": 10,
                "radiusY": 20,
                "angleOfRotation": 0,
            },
        ]

        expected_json_list = [
            {
                "type": "polygon",
                "accuracy": 0.45455,
                "parents": [],
                "children": [],
                "id": "unicorn",
                "points": [
                    {"x": 0, "y": 10},
                    {"x": 10, "y": 10},
                    {"x": 10, "y": 0},
                    {"x": 0, "y": 0},
                ],
                "area": 100,
            },
            {
                "type": "ellipse",
                "accuracy": 0.443,
                "id": "unicorn",
                "center": {"x": 769, "y": 649},
                "radiusX": 10,
                "radiusY": 20,
                "angleOfRotation": 0,
                "area": math.pi * 10 * 20,
            },
        ]

        expected_output = (
            2,
            round(math.pi * 10 * 20 + 100, 2),
            round((math.pi * 10 * 20 + 100) / 2, 2),
            round(math.pi * 10 * 20, 2),
            100.0,
        )

        result = summary_json(input_json_list, "area", do_copy=False)

        self.assertTupleEqual(result, expected_output)

        self.assertListEqual(input_json_list, expected_json_list)

    def test_summary_rounding3(self):
        input_json_list = [
            {
                "type": "polygon",
                "accuracy": 0.45455,
                "parents": [],
                "children": [],
                "id": "unicorn",
                "points": [
                    {"x": 0, "y": 10},
                    {"x": 10, "y": 10},
                    {"x": 10, "y": 0},
                    {"x": 0, "y": 0},
                ],
            },
            {
                "type": "ellipse",
                "accuracy": 0.443,
                "id": "unicorn",
                "center": {"x": 769, "y": 649},
                "radiusX": 10,
                "radiusY": 20,
                "angleOfRotation": 0,
            },
        ]

        copy_json_list = [
            {
                "type": "polygon",
                "accuracy": 0.45455,
                "parents": [],
                "children": [],
                "id": "unicorn",
                "points": [
                    {"x": 0, "y": 10},
                    {"x": 10, "y": 10},
                    {"x": 10, "y": 0},
                    {"x": 0, "y": 0},
                ],
            },
            {
                "type": "ellipse",
                "accuracy": 0.443,
                "id": "unicorn",
                "center": {"x": 769, "y": 649},
                "radiusX": 10,
                "radiusY": 20,
                "angleOfRotation": 0,
            },
        ]

        expected_output = (
            2,
            round(math.pi * 10 * 20 + 100, 3),
            round((math.pi * 10 * 20 + 100) / 2, 3),
            round(math.pi * 10 * 20, 3),
            100.0,
        )

        result = summary_json(input_json_list, "area", rounding=3)

        self.assertTupleEqual(result, expected_output)

        self.assertListEqual(input_json_list, copy_json_list)

    def test_value_error(self):
        input_json_list = [
            {
                "type": "polygon",
                "accuracy": 0.45455,
                "parents": [],
                "children": [],
                "id": "unicorn",
                "points": [
                    {"x": 0, "y": 10},
                    {"x": 10, "y": 10},
                    {"x": 10, "y": 0},
                    {"x": 0, "y": 0},
                ],
            },
            {
                "type": "ellipse",
                "accuracy": 0.443,
                "id": "unicorn",
                "center": {"x": 769, "y": 649},
                "radiusX": 10,
                "radiusY": 20,
                "angleOfRotation": 0,
            },
        ]
        self.assertRaises(ValueError, summary_json, input_json_list, "aea")



import json
import os
from tkinter import filedialog
import numpy as np

import cv2
import cytoBoom as CB

from tomni.annotation_manager import AnnotationManager
from tomni.transformers.json2contours import json2contours


# %%
json_fp = filedialog.askopenfilename(title="Select CDF JSONs.")

with open(json_fp, "rb") as f:
    dicts = json.load(f)
    print(f"Found {len(dicts)} dictionaries in json file.")


# %%
manager = AnnotationManager.from_dicts(dicts=dicts)
dicts_ = manager.to_dict(
    features=["area", "circularity", "major_axis", "average_diameter"],
    feature_multiplier=1 / 742,
    metric_unit="mm",
)
print(dicts_[0])

# %%
contours = [json2contours(d) for d in dicts]
manager = AnnotationManager.from_contours(contours=contours)

# %%
shape = (2072, 2072)
bin_mask_fp = filedialog.askopenfilename(title="Select binary mask.")
mask = cv2.imread(bin_mask_fp, cv2.IMREAD_GRAYSCALE)
manager = AnnotationManager.from_binary_mask(mask=mask)
cv2.imwrite("to_binary_mask.png", manager.to_binary_mask(shape))

# %%
shape = (2072, 2072)
labeled_mask_fp = filedialog.askopenfilename(title="Select labeled mask.")
mask = cv2.imread(labeled_mask_fp, cv2.IMREAD_GRAYSCALE)
manager = AnnotationManager.from_labeled_mask(mask=mask)
cv2.imwrite("to_labeled_mask.png", manager.to_labeled_mask(shape))

# %%
print(f"__len__: {len(manager)}")

# %%
# manager creates a generator.
count = 0
for annotation in manager:
    count += 1
print(f"Count: {count}")


# %%
dicts_ = manager.to_dict()

with open("temp.json", "w") as f:
    json.dump(dicts_, f)

# %% #simplify polygons
dicts_ = manager.to_dict(do_compress=True, epsilon=3)
with open("temp.json", "w") as f:
    json.dump(dicts_, f)


# %% to_dict with masked rois.
# define masks for a lux image, ideally you use a lux json but whatever floats your boat.

size = int(2072 / 2)
rad = int(2072 / 3)
mask_json = [
    {
        "type": "ellipse",
        "center": {"x": size, "y": size},
        "radiusX": rad,
        "radiusY": rad,
        "angleOfRotation": 0,
        "name": "A1",
    }
]

_dicts = manager.to_dict(mask_json=mask_json, min_overlap=0.9)
print(_dicts)
img_path = json_fp.replace("json", "jpg")
if os.path.exists(img_path):
    img = cv2.imread(img_path, cv2.IMREAD_COLOR)
    for _dict in _dicts:
        CB.over_draw_json(img, _dict, stroke_width=3, color=(0, 255, 0))

    CB.over_draw_json(img, mask_json[0], stroke_width=3, color=(255, 0, 0))
    cv2.imwrite("to_dict_with_mask.png", img)

# %%
# Replace with the dimensions of your image.
bin_mask = manager.to_binary_mask(shape=(2072, 2072))
cv2.imwrite("binary_mask.png", bin_mask * 255)


# %%
conts = manager.to_contours()
print(conts)


# %%
annotations = manager.filter(feature="roundness", min_val=0.5, max_val=1.0)
print(annotations)

# %%
Filter with inplace=True: manager object is updated internally. Returns manager object to allow chaining.
updated_manager = manager.filter(
    # The return does not have to be used. This is merely to show difference between inplace.
    feature="roundness",
    min_val=0.5,
    max_val=1.0,
    inplace=True,
).filter(feature="area", min_val=0, max_val=1000, inplace=True)
print(type(updated_manager))


# %%
# Filter: inplace=False returns a new list of annotations.
annotations = manager.filter(
    feature="roundness", min_val=0.5, max_val=1.0, inplace=False
)
print(type(annotations))

%%



from math import pi, sqrt
from typing import Tuple

import cv2
import numpy as np

from .. import get_center


def approximate_circle_by_area(contour: np.ndarray) -> Tuple[float, float, float]:
    """
    Approximate a circle with the same area and center as the given contour.

    Args:
        contour (np.ndarray): An OpenCV contour of a single object.

    Returns:
        Tuple[float, float, float]: A tuple containing the following values:
            - x (float): Center x position of the approximate circle.
            - y (float): Center y position of the approximate circle.
            - radius (float): Radius of the approximate circle.

    Example::

        contour = np.array([[[1, 2]], [[2, 3]], [[3, 2]], [[2, 1]]])
        center_x, center_y, circle_radius = approximate_circle_by_area(contour)
        (2.0, 2.0, 1.0)
    """
    x, y = get_center(contour)
    area = cv2.contourArea(contour)
    radius = sqrt(area / pi)
    return x, y, radius



from typing import List

from tomni.annotation_manager.annotations.point import Point


def simplify_line(line: List[Point], is_enclosed=True):
    if is_enclosed == False:
        raise NotImplemented()

    # returns empty list if `line` has exactly two elements
    if len(line) == 2:
        return line

    filtered_points = []

    # Remove point that do no add new information
    for i in range(len(line)):
        prev_p = line[i - 1]
        cur_p = line[i]
        next_p = line[i + 1] if (i + 1) < len(line) else line[0]

        """
        If ratio heigh width for prev-cur and cur-next is the same.
        The point is on a straight line
        Comparing is done with division for optimization
        """
        prev_cur_width = prev_p.x - cur_p.x
        prev_cur_height = prev_p.y - cur_p.y
        cur_next_width = cur_p.x - next_p.x
        cur_next_height = cur_p.y - next_p.y

        if prev_cur_width * cur_next_height == prev_cur_height * cur_next_width:
            continue

        filtered_points.append(cur_p)

    return filtered_points



from ..bbox_fitting import bbox_fitting_center, bbox_fitting

from .circlair_mask import make_mask_circle
from .contour_mask_maker import make_mask_contour
from .ellipse_mask import make_mask_ellipse



import cv2
import numpy as np

from ..contours2json import contours2json
from ..labels2contours import labels2contours


def get_edges(mask: np.ndarray):
    """
    Extracts and returns the edges of a binary mask image.

    This function uses the Canny edge detection algorithm to identify edges in the input binary mask image.

    Args:
        mask (np.ndarray): A binary mask represented as a NumPy array, where 1 indicates the object of interest,
            and 0 indicates the background.

    Returns:
        np.ndarray: A binary image representing the edges of the input mask. Pixels along edges are set to 1,
            while non-edge pixels are set to 0.
    """
    mask_copy = cv2.copyMakeBorder(
        mask,
        top=1,
        bottom=1,
        left=1,
        right=1,
        borderType=cv2.BORDER_CONSTANT,
        value=0,
    )
    edges = cv2.Canny(mask_copy, 50, 150)

    edges = cv2.dilate(edges, np.ones((5, 5)))

    edges = np.divide(edges, 255, dtype=np.float16)
    edges = edges.astype(np.uint8)
    edges = edges[1:-1, 1:-1]

    return edges


def mask2json(
    mask: np.ndarray,
    minimum_size_contours: int = 3,
    is_diagonal_connected=True,
    return_inner_contours: bool = False,
) -> list:
    """
    Converts a binary mask into a list of JSON objects representing contours.

    Args:
        mask (np.ndarray): A binary mask represented as a NumPy array.
        minimum_size_contours (int, optional): The minimum number of points a contour should have to be included.
            Contours with fewer points will be excluded. Defaults to 3.
        is_diagonal_connected (bool, optional): If True, diagonal pixels (e.g., [[1, 0], [0, 1]]) will be considered
            part of the same object during contour extraction. Defaults to True.
        return_inner_contours (bool, optional): If True, include internal contours (holes within the shape).
            If False, only return the external contours (outlines of the shapes). Defaults to False.

    Returns:
        list: A list of JSON objects representing contours in the mask.

    Example output::

            {
                'type': 'polygon',
                'points': [{'x': x1, 'y': y1}, {'x': x2, 'y': y2}, ...],
                'innerObjects': [
                    {
                        'type': 'polygon',
                        'points': [{'x': x3, 'y': y3}, {'x': x4, 'y': y4}, ...],
                    },
                ],
            }
    """

    connectivity = 4
    if is_diagonal_connected:
        connectivity = 8

    labels = cv2.connectedComponents(mask.astype(np.uint8), connectivity=connectivity)[
        1
    ]
    edges = get_edges(mask)

    edges = edges * labels
    del labels
    contours = labels2contours(edges, return_inner_contours=return_inner_contours)

    if return_inner_contours:
        json_objects = []
        for contour in contours:
            json_object = contours2json([contour[0]])[0]
            json_object["innerObjects"] = contours2json(contour[1])
            json_objects.append(json_object)
    else:
        json_objects = contours2json(contours)
    json_objects = [
        json_object
        for i, json_object in enumerate(json_objects)
        if len(json_objects[i]["points"]) >= minimum_size_contours
    ]

    return json_objects



from typing import Union


def check_overlap_bbox(bb1: Union[tuple, list], bb2: Union[tuple, list]) -> bool:
    """
    Checks if two bounding boxes overlap with each other. Bounding boxes are represented
    as tuples or lists in the format (xmin, ymin, xmax, ymax).

    Args:
        bb1 (Union[tuple, list]): Bounding box 1, specified as (x1, y1, x2, y2).
        bb2 (Union[tuple, list]): Bounding box 2, specified as (x1, y1, x2, y2).

    Raises:
        ValueError: If the length of either bounding box is not 4.

    Returns:
        bool: True if the bounding boxes overlap, False otherwise.

    Example::

        bb1 = (1, 1, 4, 4)
        bb2 = (3, 3, 6, 6)
        check_overlap_bbox(bb1, bb2)
        True  # The bounding boxes overlap.
    """
    if len(bb1) != 4 or len(bb2) != 4:
        raise ValueError("The bounding boxes must be tuples or list of length 4")

    if (bb1[0] >= bb2[2]) or (bb2[0] >= bb1[2]):
        return False
    elif (bb1[1] >= bb2[3]) or (bb2[1] >= bb1[3]):
        return False
    else:
        return True



import copy


def flip_json(json_object: dict, img_dim: int) -> dict:
    """
    Flip a JSON object over the y-axis.

    Args:
        json_object (dict): JSON object following the standard AxionBio format (ellipse or polygon).
        img_dim (int): The y-dimension of the image related to the JSON object.

    Raises:
        ValueError: If the JSON object type is not "ellipse" or "polygon".

    Returns:
        dict: A new JSON object with coordinates flipped over the y-axis.
    """

    newjson_object = copy.deepcopy(json_object)
    if json_object["type"] == "ellipse":
        newjson_object["center"] = flip_coordinates(json_object["center"], img_dim)

    elif json_object["type"] == "polygon":
        newjson_object["points"] = [
            flip_coordinates(point, img_dim) for point in json_object["points"]
        ]

        if "inner_points" in json_object:
            newjson_object["inner_points"] = [
                [
                    flip_coordinates(inner_point, img_dim)
                    for inner_point in inner_contour
                ]
                for inner_contour in json_object["inner_points"]
            ]

    else:
        raise ValueError(
            f"The type '{json_object['type']}' is not supported for flipping augmentation."
        )

    return newjson_object


def flip_coordinates(coordinates: dict, img_dim: int) -> dict:
    """
    Flip a dictionary representing 2D coordinates over the y-axis.

    Args:
        coordinates (dict): A dictionary representing 2D coordinates with 'x' and 'y' keys.
        img_dim (int): The y-dimension of the image.

    Returns:
        dict: A new dictionary with coordinates flipped over the y-axis.
    """
    return {"x": coordinates["x"], "y": img_dim - coordinates["y"] - 1}



import numpy as np
import cv2


def relative_difference(
    img: np.ndarray,
    gauss_size: int = 301,
    smooth_size=None,
    do_normalize: bool = False,
    resize_ratio=None,
) -> np.array:
    """
    Remove large blurry artifacts from an image by calculating the relative difference
    between the blurred and raw image.

    Args:
        img (numpy.ndarray): The input image.
        gauss_size (int, optional): The size of the Gaussian filter kernel for artifact removal.
                                    Must be an odd number. Defaults to 301.
        smooth_size (int, optional): The size of the Gaussian filter kernel applied to the image
                                    before division. Must be an odd number. Defaults to None.
        do_normalize (bool, optional): If True, the output image is normalized between the minimum
                                       and maximum values of the smoothed image. Defaults to False.
        resize_ratio (float, optional): Resize ratio for the input image before applying Gaussian blur.
                                       Used to increase processing speed. Defaults to None.

    Returns:
        numpy.ndarray: The result image after artifact removal.

    Warning:
        - The `gauss_size` should be an odd number; if it's even, it will be incremented by 1.
        - This algorithm is not fractal, meaning that applying illumination correction followed by cropping will produce a different result than cropping first.

    Note:
        This function can be used for human vision and as input for algorithms relying on relative pixel differences.
    """

    img = np.float32(img)

    if gauss_size % 2 == 0:
        gauss_size += 1

    if smooth_size:
        img = cv2.GaussianBlur(img, (smooth_size, smooth_size), 0)

    max_img = np.max(img)
    min_img = np.min(img)

    if resize_ratio:
        gb = cv2.resize(
            img,
            (0, 0),
            fx=resize_ratio,
            fy=resize_ratio,
            interpolation=cv2.INTER_LINEAR,
        )
        gb = cv2.GaussianBlur(
            gb, (gauss_size, gauss_size), 0, borderType=cv2.BORDER_REPLICATE
        )
        gb = cv2.resize(
            gb, (img.shape[1], img.shape[0]), interpolation=cv2.INTER_LINEAR
        )
    else:
        gb = cv2.GaussianBlur(img, (gauss_size, gauss_size), 0)
        gb = np.float32(gb)

    img = np.divide(img, gb + 1e-15, dtype=np.float32)

    if do_normalize:
        img = cv2.normalize(img, img, max_img, min_img, cv2.NORM_MINMAX)
    else:
        img = np.divide(img, np.max(img) + 1e-15, dtype=np.float32) * 255

    return img.astype(np.uint8)



import gc
from dataclasses import asdict
from typing import List, Tuple, Union

import cv2
import numpy as np

from tomni.annotation_manager.annotations.annotation import Annotation
from tomni.annotation_manager.annotations.point import Point
from tomni.annotation_manager.utils import overlap_object


class Ellipse(Annotation):
    def __init__(
        self,
        radius_x: float,
        center: Point,
        rotation: float,
        id: str,
        label: str = "",
        children: List[Annotation] = [],
        parents: List[Annotation] = [],
        radius_y: Union[float, None] = None,
        accuracy: float = 1,
    ):
        """Initializes a Ellipse object.

        Args:
            radius_x (float): Radius of ellipse in X direction.
            radius_y (float): Radius of ellipse in Y direction.
                Defaults to None. If None, radius_y will inherit the value of radius_x.
            center (Point): Coordinate of the center (x, y).
            rotation (float): Angle of rotation in degrees.
                Rotations will be set between 0 and 90
                If a rotation is the high or low multiplies of 180 will be added or subtracted
                If a rotation is between 90 and 180 the radii will be flipped and 90 will be subtracted
            id (str): UUID identifier.
            label (str): Class label of annotation.
            children (List[Annotation]): Tracking annotations. Refers to t+1.
            parents (List[Annotation]): Tracking annotations. Refers to t-1.
            accuracy (float, optional): The confidence of the model's prediction. Defaults to 1.
        """
        super().__init__(id, label, children, parents, accuracy)
        self._center: Point = center
        self.radius_x: float = radius_x
        if radius_y:
            self.radius_y: float = radius_y
        else:
            self.radius_y: float = radius_x
        self.rotation: float = rotation

        self._feature_multiplier: float = 1
        self._area: Union[float, None] = None
        self._aspect_ratio: Union[float, None] = None
        self._average_diameter: Union[float, None] = None
        self._circularity: Union[float, None] = None
        self._convex_hull_area: Union[float, None] = None
        self._major_axis: Union[float, None] = None
        self._minor_axis: Union[float, None] = None
        self._perimeter: Union[float, None] = None
        self._roundness: Union[float, None] = None

        self._all_features = {
            "area": {"is_ratio": False},
            "aspect_ratio": {"is_ratio": True},
            "average_diameter": {"is_ratio": False},
            "circularity": {"is_ratio": True},
            "convex_hull_area": {"is_ratio": False},
            "major_axis": {"is_ratio": False},
            "minor_axis": {"is_ratio": False},
            "perimeter": {"is_ratio": False},
            "roundness": {"is_ratio": True},
        }

    @property
    def accuracy(self):
        """Accuracy of ellipse."""
        return self._accuracy

    @accuracy.setter
    def points(self, *arg, **kwargs) -> None:
        raise SyntaxError("Accuracy is Immutable")

    @property
    def label(self):
        return super().label

    @label.setter
    def label(self, value) -> None:
        super().label = value

    @property
    def radius_x(self) -> float:
        """Radius of ellipse in X direction."""
        return self._radius_x

    @radius_x.setter
    def radius_x(self, new_radius_x):
        self._radius_x = new_radius_x

    @property
    def radius_y(self) -> float:
        """Radius of ellipse in X direction."""
        return self._radius_y

    @radius_y.setter
    def radius_y(self, new_radius_y):
        self._radius_y = new_radius_y

    @property
    def center(self) -> Point:
        """Center of ellipse."""
        return self._center

    @property
    def rotation(self) -> float:
        """Rotation in degrees."""
        return self._rotation

    @rotation.setter
    def rotation(self, new_rotation) -> float:
        # Everything above 180 is a repetition of below
        new_rotation = new_rotation % 180
        if new_rotation >= 90:
            self.radius_x, self.radius_y = self.radius_y, self.radius_x
            self._rotation = new_rotation % 90
        else:
            self._rotation = new_rotation

    @property
    def area(self) -> float:
        """Area described by pi * radius_x * radius_y.

        Returns:
            float: Ellipse's area.
        """
        if self._area is None:
            self._calculate_area()
        return self._area * self._feature_multiplier**2

    @property
    def circularity(self) -> float:
        """Circularity described by 4 * pi * Area / Perimeter**2.

        Returns:
            float: Ellipse's circularity.
        """
        if self._circularity is None:
            self._calculate_circularity()
        return self._circularity

    @property
    def convex_hull_area(self) -> Union[float, None]:
        """Calculate the convex hull area of an ellipse.

        This property computes the area of the convex hull of the ellipse using the formula:
        Area = π * radius_x * radius_y.

        Returns:
            Union[float, None]: The convex hull area of the ellipse, or None if it cannot be calculated.
        """
        if self._convex_hull_area is None:
            self._calculate_convex_hull_area()
        return self._convex_hull_area * self._feature_multiplier**2

    @property
    def average_diameter(self) -> float:
        """Returns the average diameter of an ellipse.

        Returns:
            float: Average diameter.
        """
        if self._average_diameter is None:
            self._calculate_average_diameter()
        return self._average_diameter * self._feature_multiplier

    @property
    def minor_axis(self) -> float:
        """Returns the minor axis of an ellipse.

        Returns:
            float: Minor axis length.
        """
        if self._minor_axis is None:
            self._calculate_minor_axis()
        return self._minor_axis * self._feature_multiplier

    @property
    def major_axis(self) -> float:
        """Return the major axis of the polygon.

        Returns:
            float: major axis length.
        """
        if self._major_axis is None:
            self._calculate_major_axis()
        return self._major_axis * self._feature_multiplier

    @property
    def aspect_ratio(self) -> float:
        """Ratio between minor and major axis.

        Returns:
            float: Ellipse's aspect ratio.
        """
        if self._aspect_ratio is None:
            self._calculate_aspect_ratio()
        return self._aspect_ratio

    @property
    def perimeter(self) -> float:
        """Perimeter described by 2*pi*sqrt((radius_x**2 + radius_y**2) / 2).

        Returns:
            float: Ellipse's perimeter.
        """
        if self._perimeter is None:
            self._calculate_perimeter()
        return self._perimeter * self._feature_multiplier

    @property
    def roundness(self) -> Union[float, None]:
        """Roundness: Area / (radius_enclosing_circle**2 * pi).

        Returns:
            Union[float, None]: Polygon's roundness in [0, 1] or None.
        """

        if self._roundness is None:
            self._calculate_roundness()
        return self._roundness

    def to_dict(
        self,
        decimals: int = 2,
        features: Union[List[str], None] = None,
        metric_unit="",
        feature_multiplier: int = 1,
        **kwargs,
    ) -> dict:
        """Returns a dictionary of the annotation in AxionBio format.

        Args:
            decimals (int, optional): The number of decimals to use when rounding. Defaults to 2.
            features (Union[List[str], None], optional): The features that are calculated and returned in the dict.
                Defaults to None, which means all features are calculated and returned.
            metric_unit (str, optional): The suffic added to the dict key names in camelCasing. Defaults to "".
                For example: "area" with suffix "um" becomes "areaUm"
            feature_multiplier (int, optional): A multiplier used during feature calculations. For example; 1/742. Defaults to 1.

        Raises:
            ValueError: It raises and error if any of the features are not compatible with the Annotation Manager.

        Returns:
            dict: a dictionary of the annotation in AxionBio format with the calculated features.
        """
        self._feature_multiplier = feature_multiplier

        # Check if features list is provided; otherwise, return all features
        if features is None:
            features = list(self._all_features.keys())
        else:
            missing_features = set(features).difference(set(self._all_features.keys()))
            if missing_features:
                raise ValueError(
                    f"The following features are not compatible with the Annotation Manager: {', '.join(missing_features)}"
                )

        dict_ellipse = {
            "type": "ellipse",
            "center": asdict(self.center),
            "radiusX": self.radius_x,
            "radiusY": self.radius_y,
            "angleOfRotation": self.rotation,
        }

        for feature in features:
            feature_name = (
                feature
                if self._all_features[feature]["is_ratio"]
                else feature + "_" + metric_unit
            )

            # Convert snake_casing to camelCasing
            first_word, *remaining_words = feature_name.split("_")
            feature_name = "".join(
                [first_word.lower(), *map(str.title, remaining_words)]
            )

            dict_ellipse[feature_name] = round(getattr(self, feature), decimals)

        super_dict = super().to_dict(decimals=decimals)
        dict_return_value = {**super_dict, **dict_ellipse}
        return dict_return_value

    def to_binary_mask(self, shape: Tuple[int, int]) -> np.ndarray:
        """Transform an ellipse to a binary mask.

        Args:
            shape (Tuple[int, int]): Shape of the new ellipse's binary mask.

        Returns:
            np.ndarray: A binary mask in [0, 1].
        """
        mask = np.zeros(shape, dtype=np.uint8)
        return cv2.ellipse(
            mask,
            center=(self.center.x, self.center.y),
            axes=(self.radius_x, self.radius_y),
            angle=self.rotation,
            startAngle=0,
            endAngle=360,
            color=1,
            thickness=-1,
        )

    def is_in_mask(self, mask_json: List[dict], min_overlap: float = 0.9) -> bool:
        """Check if an ellipse is within a binary mask.

        Args:
            mask_json (List[dict]): A list of dict masks in AxionBio dict format.
            min_overlap (float, optional): Minimum overlap required between the ellipse and the mask, expressed as a value between 0 and 1. Defaults to 0.9.

        Returns:
            bool: True if the ellipse is within a mask and meets the required overlap, False otherwise.
        """

        json_object = {
            "type": "ellipse",
            "center": {"x": self.center.x, "y": self.center.y},
            "radiusX": self.radius_x,
            "radiusY": self.radius_y,
            "angleOfRotation": self.rotation,
        }

        for mask in mask_json:
            overlap_ratio = overlap_object(json_object, mask)
            if overlap_ratio >= min_overlap:
                return True

        return False

    def _calculate_circularity(self) -> None:
        if not self._area:
            self._calculate_area()

        if not self._perimeter:
            self._calculate_perimeter()

        self._circularity = 4 * np.pi * self._area / self._perimeter**2

    def _calculate_perimeter(self) -> None:
        self._perimeter = (
            2 * np.pi * np.sqrt((self._radius_x**2 + self._radius_y**2) / 2)
        )

    def _calculate_roundness(self) -> None:
        if not self._area:
            self._calculate_area()

        if not self._major_axis:
            self._calculate_major_axis()

        if self.major_axis == 0:
            self._roundness = 0
        else:
            enclosing_circle_area = (self._major_axis / 2) ** 2 * np.pi
            self._roundness = self._area / enclosing_circle_area

    def _calculate_area(self) -> None:
        self._area = np.pi * self._radius_x * self._radius_y

    def _calculate_convex_hull_area(self) -> None:
        if not self._area:
            self._calculate_area()
        self._convex_hull_area = self._area

    def _calculate_minor_axis(self) -> None:
        self._minor_axis = min(self._radius_x, self._radius_y) * 2

    def _calculate_major_axis(self) -> None:
        self._major_axis = max(self._radius_x, self._radius_y) * 2

    def _calculate_average_diameter(self) -> None:
        self._average_diameter = (self._radius_x * 2 + self._radius_y * 2) / 2

    def _calculate_aspect_ratio(self) -> None:
        if self._radius_x == self._radius_y:
            self._aspect_ratio = 1.0
        elif self._radius_x > self._radius_y:
            self._aspect_ratio = self._radius_y / self._radius_x
        else:
            self._aspect_ratio = self._radius_x / self._radius_y

    def __eq__(self, other):
        if self.radius_x != other.radius_x:
            return False

        if self.radius_y != other.radius_y:
            return False

        if self.rotation != other.rotation:
            return False

        if self.center != other.center:
            return False

        return True



import numpy as np
from typing import Tuple


def fit_rect_around_ellipse(
    x: int, y: int, r1: float, r2: float, alpha: float
) -> Tuple[int, int, int, int]:
    """
    Compute the bounding box coordinates of an ellipse given its parameters.

    Args:
        x (int): The x-coordinate of the center of the ellipse.
        y (int): The y-coordinate of the center of the ellipse.
        r1 (float): The longest radius of the ellipse.
        r2 (float): The shortest radius of the ellipse, which is 90 degrees away from r1.
        alpha (float): The degrees between r1 and x-axis.

    Returns:
        Tuple[int, int, int, int]: A tuple containing the coordinates of the bounding box in the format (x1, y1, x2, y2).

    Special thanks to: https://math.stackexchange.com/questions/91132/how-to-get-the-limits-of-rotated-ellipse
    """
    alpha = alpha / 180 * np.pi
    a = np.sqrt(r1**2 * np.cos(alpha) ** 2 + r2**2 * np.sin(alpha) ** 2)
    b = np.sqrt(r1**2 * np.sin(alpha) ** 2 + r2**2 * np.cos(alpha) ** 2)
    x1 = int(round(x - a))
    y1 = int(round(y - b))
    x2 = int(round(x + a))
    y2 = int(round(y + b))
    return x1, y1, x2, y2



from __future__ import absolute_import
from unittest import TestCase
from . import bbox_fitting
import numpy as np


class TestBbox_fitting(TestCase):
    def test_raise_warning_image(self):
        input_array = [
            [0, 1, 1, 1, 0],
            [1, 1, 1, 1, 1],
            [1, 1, 1, 2, 1],
            [1, 1, 1, 3, 1],
            [0, 1, 1, 1, 0],
        ]
        with self.assertRaises(TypeError):
            result = bbox_fitting(input_array, 1, 1, 3, 3)

    def test_raise_warning_float_cor(self):
        input_array = np.array(
            [
                [0, 1, 1, 1, 0],
                [1, 1, 1, 1, 1],
                [1, 1, 1, 2, 1],
                [1, 1, 1, 3, 1],
                [0, 1, 1, 1, 0],
            ]
        )
        with self.assertRaises(ValueError):
            result = bbox_fitting(input_array, 1.1, 1, 3, 3)
        with self.assertRaises(ValueError):
            result = bbox_fitting(input_array, 1, 1.3, 3, 3)
        with self.assertRaises(ValueError):
            result = bbox_fitting(input_array, 1, 1, 3.9, 3)
        with self.assertRaises(ValueError):
            result = bbox_fitting(input_array, 1, 1, 3, 3.0)

    def test_crop_5x5_3x3(self):
        input_array = np.array(
            [
                [0, 1, 1, 1, 0],
                [1, 1, 1, 1, 1],
                [1, 1, 1, 2, 1],
                [1, 1, 1, 3, 1],
                [0, 1, 1, 1, 0],
            ]
        )

        expected = np.array(
            [
                [1, 1, 1],
                [1, 1, 2],
                [1, 1, 3],
            ]
        )

        result = bbox_fitting(input_array, 1, 1, 4, 4)
        np.testing.assert_array_equal(expected, result)

    def test_crop_7x5_3x3(self):
        input_array = np.array(
            [
                [0, 1, 0, 1, 0],
                [0, 1, 1, 1, 0],
                [1, 1, 4, 1, 1],
                [1, 1, 5, 1, 1],
                [1, 1, 1, 1, 1],
                [0, 1, 1, 1, 0],
                [0, 1, 0, 1, 0],
            ]
        )

        expected = np.array(
            [
                [1, 4, 1],
                [1, 5, 1],
                [1, 1, 1],
            ]
        )

        result = bbox_fitting(input_array, 1, 2, 4, 5)
        np.testing.assert_array_equal(expected, result)

    def test_padding_3x3_5x5(self):
        input_array = np.array(
            [
                [5, 1, 1],
                [1, 1, 2],
                [1, 1, 3],
            ]
        )

        expected = np.array(
            [
                [0, 0, 0, 0, 0],
                [0, 5, 1, 1, 0],
                [0, 1, 1, 2, 0],
                [0, 1, 1, 3, 0],
                [0, 0, 0, 0, 0],
            ]
        )

        result = bbox_fitting(input_array, -1, -1, 4, 4)
        np.testing.assert_array_equal(expected, result)

    def test_padding_3x3_5x7(self):
        input_array = np.array(
            [
                [5, 1, 1],
                [1, 1, 2],
                [1, 1, 3],
            ]
        )

        expected = np.array(
            [
                [0, 0, 0, 0, 0],
                [0, 5, 1, 1, 0],
                [0, 1, 1, 2, 0],
                [0, 1, 1, 3, 0],
                [0, 0, 0, 0, 0],
                [0, 0, 0, 0, 0],
                [0, 0, 0, 0, 0],
            ]
        )

        result = bbox_fitting(input_array, -1, -1, 4, 6)
        np.testing.assert_array_equal(expected, result)

    def test_padding_and_cropping_3x3_3x4(self):
        input_array = np.array(
            [
                [5, 1, 1],
                [1, 1, 2],
                [1, 1, 3],
            ]
        )

        expected = np.array([[1, 1, 0], [1, 2, 0], [1, 3, 0], [0, 0, 0]])

        result = bbox_fitting(input_array, 1, 0, 4, 4)
        np.testing.assert_array_equal(expected, result)

    def test_padding_and_cropping_3x3_3x4_color(self):
        input_array = np.array(
            [
                [[5, 4, 3], [1, 9, 2], [1, 8, 2]],
                [[1, 2, 2], [1, 1, 2], [1, 7, 2]],
                [[1, 4, 2], [1, 5, 2], [1, 6, 2]],
            ]
        )

        expected = np.array(
            [
                [[1, 9, 2], [1, 8, 2], [0, 0, 0]],
                [[1, 1, 2], [1, 7, 2], [0, 0, 0]],
                [[1, 5, 2], [1, 6, 2], [0, 0, 0]],
                [[0, 0, 0], [0, 0, 0], [0, 0, 0]],
            ]
        )

        result = bbox_fitting(input_array, 1, 0, 4, 4)
        np.testing.assert_array_equal(expected, result)

    def test_underflow_left(self):
        input_array = np.array(
            [
                [5, 1, 1],
                [1, 1, 2],
                [1, 1, 3],
            ]
        )

        expected = np.array(
            [[0, 0, 0, 0], [0, 0, 0, 0], [0, 0, 0, 0], [0, 0, 0, 0], [0, 0, 0, 0]]
        )

        result = bbox_fitting(input_array, -5, 1, -1, 6)
        np.testing.assert_array_equal(expected, result)

    def test_overflow_right(self):
        input_array = np.array(
            [
                [5, 1, 1],
                [1, 1, 2],
                [1, 1, 3],
            ]
        )

        expected = np.array(
            [[0, 0, 0, 0], [0, 0, 0, 0], [0, 0, 0, 0], [0, 0, 0, 0], [0, 0, 0, 0]]
        )

        result = bbox_fitting(input_array, 3, 1, 7, 6)
        np.testing.assert_array_equal(expected, result)

    def test_overflow_down(self):
        input_array = np.array(
            [
                [5, 1, 1],
                [1, 1, 2],
                [1, 1, 3],
            ]
        )

        expected = np.array([[0, 0, 0], [0, 0, 0], [0, 0, 0], [0, 0, 0]])

        result = bbox_fitting(input_array, 1, 9, 4, 13)
        np.testing.assert_array_equal(expected, result)

    def test_underflow_up(self):
        input_array = np.array(
            [
                [5, 1, 1],
                [1, 1, 2],
                [1, 1, 3],
            ]
        )

        expected = np.array(
            [[0, 0, 0, 0], [0, 0, 0, 0], [0, 0, 0, 0], [0, 0, 0, 0], [0, 0, 0, 0]]
        )

        result = bbox_fitting(input_array, 1, -7, 5, -2)
        np.testing.assert_array_equal(expected, result)

    def test_padding_3x3_5x7_different_color(self):
        """
        Don't argure about how color is writen. It is color without, because I don't like u.
        """
        input_array = np.array(
            [
                [5, 1, 1],
                [1, 1, 2],
                [1, 1, 3],
            ]
        )

        expected = np.array(
            [
                [255, 255, 255, 255, 255],
                [255, 5, 1, 1, 255],
                [255, 1, 1, 2, 255],
                [255, 1, 1, 3, 255],
                [255, 255, 255, 255, 255],
                [255, 255, 255, 255, 255],
                [255, 255, 255, 255, 255],
            ]
        )

        result = bbox_fitting(input_array, -1, -1, 4, 6, padding_value=255)
        np.testing.assert_array_equal(expected, result)



import numpy as np


def img_paste(big_img: np.ndarray, small_img: np.ndarray, x: int, y: int) -> None:
    """
    Paste a smaller image onto a larger image at a specified position (image coordinates).

    Args:
        big_img (np.ndarray): The larger image to which the smaller image will be added.
        small_img (np.ndarray): The smaller image to be pasted onto the larger image.
        x (int): The x-coordinate position in the larger image where the top-left corner of the smaller image will be placed.
        y (int): The y-coordinate position in the larger image where the top-left corner of the smaller image will be placed.

    Returns:
        None

    Note:
        The function checks if the smaller image can fit within the boundaries of the larger image at the specified position (x, y). If it can fit, the smaller image is pasted onto the larger image.
    """

    size_bimg = (big_img.shape[1], big_img.shape[0])
    size_simg = (small_img.shape[1], small_img.shape[0])

    # calculate position in big image
    # START
    if x < 0:
        start_x = 0
        crop_x_start = -x
    elif x > size_bimg[0] - 1:
        return
    else:
        start_x = x
        crop_x_start = 0

    if y < 0:
        start_y = 0
        crop_y_start = -y
    elif y > size_bimg[1] - 1:
        return
    else:
        start_y = y
        crop_y_start = 0

    # STOP
    if x + size_simg[0] > size_bimg[0]:
        stop_x = size_bimg[0]
        if x >= 0:
            crop_x_stop = size_bimg[0] - x
        else:
            crop_x_stop = size_bimg[0] + crop_x_start

    elif x + size_simg[0] < 0:
        return
    else:
        stop_x = x + size_simg[0]
        crop_x_stop = size_simg[0]

    if y + size_simg[1] > size_bimg[1]:
        stop_y = size_bimg[1]
        if y >= 0:
            crop_y_stop = size_bimg[1] - y
        else:
            crop_y_stop = size_bimg[1] + crop_y_start
    elif y + size_simg[1] < 0:
        return
    else:
        stop_y = y + size_simg[1]
        crop_y_stop = size_simg[1]

    # Crop smaller image to fit
    overlay_img = small_img[crop_y_start:crop_y_stop, crop_x_start:crop_x_stop]
    big_img[start_y:stop_y, start_x:stop_x] = overlay_img



import numpy as np
from scipy.sparse import csr_matrix, find
import numpy_indexed as npi


def labels2listsOfPoints(labels: np.ndarray) -> np.ndarray:
    """
    Convert labeled regions into lists of points.

    Args:
        labels (np.ndarray): An array where every pixel is labeled to indicate which object it belongs to.
            The labels should be non-negative integers.

    Returns:
        List[List[Tuple[int, int]]]: A list of lists, where each inner list contains the (x, y) coordinates
            of points belonging to the corresponding label. The list is organized such that
            `output[label]` contains the points for the label with value `label`. The label values must start
            from 0 and be contiguous integers.

    Raises:
        ValueError: If the number of labels exceeds the supported limit.

    Note:
        - The function automatically determines the appropriate data type for labels based on their maximum value.
        - It uses sparse matrix operations to efficiently extract points for each label.
    """

    if np.max(labels) < 256:
        dataType = np.uint8
    elif np.max(labels) < 65536:
        dataType = np.uint16
    elif np.max(labels) < 4294967296:
        dataType = np.uint32
    else:
        raise ValueError("There are to many labels")

    # Memory check number of labels
    c = csr_matrix(labels, dtype=dataType)
    x, y, v = find(c)
    g = npi.group_by(v)
    output = g.split_array_as_list(np.array([y, x]).transpose())
    return output



from __future__ import division
import numpy as np
import cv2
from . import bbox_fitting_center


def make_mask_circle(img_shape: tuple, diameter: int) -> np.ndarray:
    """
    Create a boolean image with a circle of True surrounded by False.

    Args:
        img_shape (tuple): The shape of the image given in image coordinates, represented as (width, height).
        diameter (int): The diameter of the circle in pixels.

    Returns:
        np.ndarray: A boolean image where the area enclosed by the circle is True and the rest is False.

    Note:
        - The function uses either 'make_small_mask_circle' or 'make_big_mask_circle' based on the size of the circle.
        - In case the diameter is greater than or equal to 100 pixels, the 'make_big_mask_circle' function is used for memory efficiency.
    """
    if diameter < 100:
        return make_small_mask_circle(img_shape, diameter)
    else:
        return make_big_mask_circle(img_shape, diameter)


def make_small_mask_circle(img_shape, diameter):
    """
    Create a boolean image with a small circle of True surrounded by False.

    Args:
        img_shape (tuple): The shape of the image given in image coordinates, represented as (width, height).
        diameter (int): The diameter of the circle in pixels.

    Returns:
        np.ndarray: A boolean image with a small circle represented as True values surrounded by False values.

    """
    xx, yy = np.mgrid[: img_shape[1], : img_shape[0]] * 2

    xx = xx.astype(np.int32)
    yy = yy.astype(np.int32)

    xx -= img_shape[1] - 1
    yy -= img_shape[0] - 1

    circle = xx**2 + yy**2
    kernel = circle <= diameter**2

    return kernel


def make_big_mask_circle(img_shape, diameter):
    """
    Create a boolean image with a large circle of True surrounded by False.

    Args:
        img_shape (tuple): The shape of the image given in image coordinates, represented as (width, height).
        diameter (int): The diameter of the circle in pixels.

    Returns:
        np.ndarray: A boolean image with a large circle represented as True values surrounded by False values.

    """
    out = cv2.getStructuringElement(cv2.MORPH_ELLIPSE, (diameter, diameter))
    out = bbox_fitting_center(out, img_shape)
    return out.astype(bool)



from typing import List


def json2dict(
    json_list: List[dict], keywords: list = ["type", "area", "center", "id"]
) -> dict:
    """
    Transform a list of JSON objects in standard AxionBio format into a dictionary
    using a list of specified keywords. The resulting dictionary contains lists of values
    extracted from the JSON objects for each keyword, as well as an "index" list with object indexes.

    Args:
        json_list (List[dict]): A list of JSON objects in standard AxionBio format.
        keywords (List[str], optional): A list of keywords to extract from the JSON objects.
            Defaults to ["type", "area", "center", "id"].

    Returns:
        Dict[str, list]: A dictionary containing lists of values for each keyword extracted
            from the JSON objects, including an "index" list with object indexes.
    """

    keywords.append("index")
    json_dict = {}

    for key in keywords:
        json_dict[key] = []

    for i, json_object in enumerate(json_list):
        for key in keywords:
            if key == "index":
                json_dict[key].append(i)
            elif key not in json_object.keys() and key != "index":
                json_dict[key].append(None)
            else:
                if isinstance(json_object[key], dict):
                    variable = [json_object[key][x] for x in json_object[key].keys()]
                else:
                    variable = json_object[key]
                json_dict[key].append(variable)

    return json_dict



from __future__ import division
import numpy as np
import cv2
from . import bbox_fitting


def make_mask_ellipse(image_size, x1, y1, rx, ry):
    """
    Create an image with an ellipse of True surrounded by False.

    Args:
        image_size (tuple): The size of the image (width, height).
        x1 (int): The x-coordinate of the center of the ellipse.
        y1 (int): The y-coordinate of the center of the ellipse.
        rx (int): The length of the radius on the x-axis of the ellipse.
        ry (int): The length of the radius on the y-axis of the ellipse.

    Returns:
        np.ndarray: An image with an ellipse represented as True values surrounded by False values.

    Note:
        - The function uses either 'make_small_mask_ellipse' or 'make_big_mask_ellipse' based on the size of the ellipse.
        - For rotated ellipses, additional support for 'alpha' can be added in future updates.

    """
    x1 = int(x1)
    y1 = int(y1)
    rx = int(rx)
    ry = int(ry)

    if rx < 1 or ry < 1:
        raise ValueError("Radii must be greater than 1.")
    else:
        if (rx < 100) and (ry < 100):
            return make_small_mask_ellipse(image_size, x1, y1, rx, ry)
        else:
            return make_big_mask_ellipse(image_size, x1, y1, rx, ry)


def make_small_mask_ellipse(image_size, x1, y1, rx, ry):
    """
    Create an image with a small ellipse of True surrounded by False.

    Args:
        image_size (tuple): The size of the image (width, height).
        x1 (int): The x-coordinate of the center of the ellipse.
        y1 (int): The y-coordinate of the center of the ellipse.
        rx (int): The length of the radius on the x-axis of the ellipse.
        ry (int): The length of the radius on the y-axis of the ellipse.

    Returns:
        np.ndarray: An image with a small ellipse represented as True values surrounded by False values.

    """
    yy, xx = np.mgrid[: image_size[1], : image_size[0]]

    xx = xx.astype(np.int32)
    yy = yy.astype(np.int32)

    # Center point of the ellipse becomes the (0, 0) point of the image
    xx -= x1
    yy -= y1

    # The ellipse formula used to calculate the points is
    # x^2 * (ry)^2 + y^2 * (rx)^2
    # Calculating the value of each point depending on the ellipse equation
    ellipse = (xx**2) * (ry**2) + (yy**2) * (rx**2)

    # The ellipse formula determines if a point is inside the ellipse or not
    # x^2 * (ry)^2 + y^2 * (rx)^2 <= (rx)^2 * (ry)^2
    kernel = ellipse <= (rx**2) * (ry**2)
    kernel = kernel * (np.abs(xx) <= rx)
    kernel = kernel * (np.abs(yy) <= ry)
    return kernel


def make_big_mask_ellipse(image_size, xe, ye, rex, rey):
    """
    Create an image with a large ellipse of True surrounded by False.

    Args:
        image_size (tuple): The size of the image (width, height).
        xe (int): The x-coordinate of the ellipse.
        ye (int): The y-coordinate of the ellipse.
        rex (int): The radius on the x-axis of the ellipse.
        rey (int): The radius on the y-axis of the ellipse.

    Returns:
        np.ndarray: An image with a large ellipse represented as True values surrounded by False values.

    """

    out = cv2.getStructuringElement(cv2.MORPH_ELLIPSE, (rex * 2, rey * 2))
    x0 = -(xe - rex)
    y0 = -(ye - rey)
    x1 = image_size[0] + x0
    y1 = image_size[1] + y0
    out = bbox_fitting(out, x0, y0, x1, y1)
    return out



from __future__ import absolute_import, division
from typing import List
from . import bbox_fitting
from math import ceil
import numpy as np


def bbox_fitting_center(img: np.ndarray, size, padding_value: int = 0) -> np.ndarray:
    """
    Creates an image of size 'size' in IMAGE-Coordinates by centering and padding the original image.

    Args:
        img (numpy.ndarray): The original image.
        size (List[int]): The target size of the output image as [x, y].
        padding_value (int, optional): The value used for padding. Defaults to 0.

    Returns:
        numpy.ndarray: The resulting image of the specified size.

    Raises:
        TypeError: If img is not a numpy.ndarray.
        ValueError: If size elements are not positive integers.

    Example::

        img = np.zeros((100, 100))
        size = [50, 50]
        result = bbox_fitting_center(img, size)
        result.shape
        (50, 50)  # The resulting image size.
    """
    if not isinstance(img, np.ndarray):
        raise TypeError("Img needs to be a numpy.ndarry not {}".format(type(img)))

    if not isinstance(size[0], int) or not isinstance(size[1], int):
        raise ValueError(
            "x1, y1, x2 and y2 need to be a positive int not {}, {}".format(
                type(size[0]), type(size[1])
            )
        )
    x1 = -(size[0] - img.shape[1]) / 2
    y1 = -(size[1] - img.shape[0]) / 2
    x2 = x1 + size[0]
    y2 = y1 + size[1]

    # Round the number to fit exactly with the pixels
    x1 = int(ceil(x1))
    y1 = int(ceil(y1))
    x2 = int(ceil(x2))
    y2 = int(ceil(y2))

    return bbox_fitting(img, x1=x1, y1=y1, x2=x2, y2=y2, padding_value=padding_value)



import cv2
import numpy as np
from typing import List


def json2labels(json_list: List[dict], output_dim: tuple) -> np.ndarray:
    """
    Convert a list of JSON objects representing objects to a labeled segmentation map.

    Args:
        json_list (list of dict): A list of objects in JSON format, each describing an object to be labeled.
        output_dim (tuple of int): The dimensions of the output image in image coordinates, specified as (width, height).

    Returns:
        np.ndarray: A segmentation map where each object has a unique label.

    Raises:
        ValueError: If an unsupported object type is encountered in the JSON list.
    """

    seg_map = np.zeros((output_dim[1], output_dim[0]), dtype=np.uint8)

    for i, obj in enumerate(json_list):
        if obj["type"] == "ellipse":
            cv2.ellipse(
                seg_map,
                (obj["center"]["x"], obj["center"]["y"]),
                (obj["radiusX"], obj["radiusY"]),
                obj["angleOfRotation"],
                0,
                360,
                color=i + 1,
                thickness=-1,
            )
        elif obj["type"] == "polygon":
            points = np.empty((len(obj["points"]), 2))
            for j in range(len(obj["points"])):
                points[j,] = [obj["points"][j]["x"], obj["points"][j]["y"]]
            cv2.fillConvexPoly(seg_map, np.int32([points]), i + 1)
        else:
            raise ValueError("Object type not supported")

    return seg_map



from __future__ import absolute_import
from unittest import TestCase
from .main import vgg2json


class TestVGG2Json(TestCase):
    def test_select_vgg_2_Json_one_file(self):
        input_vgg_json = {
            "test1.jpg": {
                "filename": "test1.jpg",
                "size": 151409,
                "regions": [
                    {
                        "shape_attributes": {
                            "name": "polygon",
                            "all_points_x": [
                                505,
                                478,
                            ],
                            "all_points_y": [
                                745,
                                756,
                            ],
                        },
                        "region_attributes": {"name": "organoid"},
                    },
                    {
                        "shape_attributes": {
                            "name": "ellipse",
                            "cx": 769,
                            "cy": 649,
                            "rx": 24,
                            "ry": 23,
                            "theta": 0,
                        },
                        "region_attributes": {},
                    },
                ],
            },
        }

        expected_output = [
            [
                {
                    "type": "polygon",
                    "parents": [],
                    "children": [],
                    "id": "unicorn",
                    "points": [
                        {"x": 505, "y": 745},
                        {"x": 478, "y": 756},
                    ],
                },
                {
                    "type": "ellipse",
                    "id": "unicorn",
                    "center": {"x": 769, "y": 649},
                    "radiusX": 24,
                    "radiusY": 23,
                    "angleOfRotation": 0,
                },
            ]
        ]

        result = vgg2json(input_vgg_json)
        result[0][0]["id"] = "unicorn"
        result[0][1]["id"] = "unicorn"

        self.assertEqual(len(result), len(expected_output))
        for i in range(len(result)):
            self.assertDictEqual(result[i][0], expected_output[i][0])
            self.assertDictEqual(result[i][1], expected_output[i][1])

    def test_circle(self):
        input_vgg_json = {
            "test1.jpg": {
                "filename": "test1.jpg",
                "size": 151409,
                "regions": [
                    {
                        "shape_attributes": {
                            "name": "polygon",
                            "all_points_x": [
                                505,
                                478,
                            ],
                            "all_points_y": [
                                745,
                                756,
                            ],
                        },
                        "region_attributes": {"name": "organoid"},
                    },
                    {
                        "shape_attributes": {
                            "name": "circle",
                            "cx": 769,
                            "cy": 649,
                            "r": 24,
                        },
                        "region_attributes": {},
                    },
                ],
            },
        }

        expected_output = [
            [
                {
                    "type": "polygon",
                    "parents": [],
                    "children": [],
                    "id": "unicorn",
                    "points": [
                        {"x": 505, "y": 745},
                        {"x": 478, "y": 756},
                    ],
                },
                {
                    "type": "ellipse",
                    "id": "unicorn",
                    "center": {"x": 769, "y": 649},
                    "radiusX": 24,
                    "radiusY": 24,
                    "angleOfRotation": 0,
                },
            ]
        ]

        result = vgg2json(input_vgg_json)
        result[0][0]["id"] = "unicorn"
        result[0][1]["id"] = "unicorn"

        self.assertEqual(len(result), len(expected_output))
        for i in range(len(result)):
            self.assertDictEqual(result[i][0], expected_output[i][0])
            self.assertDictEqual(result[i][1], expected_output[i][1])

    def test_select_vgg_2_Json_multi_file(self):
        input_vgg_json = {
            "test1.jpg": {
                "filename": "test1.jpg",
                "size": 151409,
                "regions": [
                    {
                        "shape_attributes": {
                            "name": "polygon",
                            "all_points_x": [
                                505,
                                478,
                            ],
                            "all_points_y": [
                                745,
                                756,
                            ],
                        },
                        "region_attributes": {"name": "organoid"},
                    },
                ],
            },
            "test2.jpg": {
                "filename": "test2.jpg",
                "size": 151409,
                "regions": [
                    {
                        "shape_attributes": {
                            "name": "polygon",
                            "all_points_x": [
                                502,
                                472,
                            ],
                            "all_points_y": [
                                742,
                                752,
                            ],
                        },
                        "region_attributes": {"name": "organoid"},
                    },
                ],
            },
        }

        expected_output = [
            [
                {
                    "type": "polygon",
                    "parents": [],
                    "children": [],
                    "id": "unicorn",
                    "points": [
                        {"x": 505, "y": 745},
                        {"x": 478, "y": 756},
                    ],
                },
            ],
            [
                {
                    "type": "polygon",
                    "parents": [],
                    "children": [],
                    "id": "unicorn",
                    "points": [
                        {"x": 502, "y": 742},
                        {"x": 472, "y": 752},
                    ],
                },
            ],
        ]

        result = vgg2json(input_vgg_json)
        result[0][0]["id"] = "unicorn"
        result[1][0]["id"] = "unicorn"

        self.assertEqual(len(result), len(expected_output))
        for i in range(len(result)):
            for j in range(len(result[i])):
                self.assertDictEqual(result[i][j], expected_output[i][j])



from unittest import TestCase

from tomni.annotation_manager.annotations.point import Point

from .main import simplify_line


class TestPolygonSimplify(TestCase):
    def test_positive_squares_extra_points(self):
        line = [
            Point(1, 1),
            Point(1, 30),
            Point(15, 30),
            Point(30, 30),
            Point(30, 14),
            Point(30, 1),
        ]
        expected = [Point(1, 1), Point(1, 30), Point(30, 30), Point(30, 1)]
        result = simplify_line(line)

        self.assertListEqual(expected, result)

    def test_remove_first_and_last(self):
        line = [
            Point(10, 1),
            Point(1, 1),
            Point(1, 30),
            Point(15, 30),
            Point(30, 30),
            Point(30, 14),
            Point(30, 1),
            Point(20, 1),
        ]
        expected = [Point(1, 1), Point(1, 30), Point(30, 30), Point(30, 1)]
        result = simplify_line(line)

        self.assertListEqual(expected, result)

    def test_very_small_5_points(self):
        line = [
            Point(1, 1),
            Point(1, 30),
            Point(15, 30.000001),
            Point(30, 30),
            Point(30, 1),
        ]
        expected = [
            Point(1, 1),
            Point(1, 30),
            Point(15, 30.000001),
            Point(30, 30),
            Point(30, 1),
        ]
        result = simplify_line(line)

        self.assertListEqual(expected, result)

    def test_triangle(self):
        line = [
            Point(1, 5),
            Point(3, 1),
            Point(5, 5),
        ]
        expected = [
            Point(1, 5),
            Point(3, 1),
            Point(5, 5),
        ]
        result = simplify_line(line)

        self.assertListEqual(expected, result)

    def test_triangle_2(self):
        line = [
            Point(1, 5),
            Point(2, 3),
            Point(3, 1),
            Point(5, 5),
        ]
        expected = [
            Point(1, 5),
            Point(3, 1),
            Point(5, 5),
        ]
        result = simplify_line(line)

        self.assertListEqual(expected, result)

    def test_triangle_3(self):
        line = [
            Point(1, 5),
            Point(2, 3.000001),
            Point(3, 1),
            Point(5, 5),
        ]
        expected = [
            Point(1, 5),
            Point(2, 3.000001),
            Point(3, 1),
            Point(5, 5),
        ]
        result = simplify_line(line)

        self.assertListEqual(expected, result)

    def test_two_points(self):
        line = [
            Point(1, 1),
            Point(5, 5),
        ]
        expected = [
            Point(1, 1),
            Point(5, 5),
        ]
        actual = simplify_line(line)

        self.assertListEqual(expected, actual)
